from django.db import models
from django.utils.translation import gettext_lazy as _


class FloorChoicesEnum(models.TextChoices):
    BASEMENT = "BASEMENT", _("Basement")
    GROUND_FLOOR = "GROUND_FLOOR", _("Ground Floor")
    FIRST_FLOOR = "FIRST_FLOOR", _("First Floor")
    SECOND_FLOOR = "SECOND_FLOOR", _("Second Floor")
    THIRD_FLOOR = "THIRD_FLOOR", _("Third Floor")
    FOURTH_FLOOR = "FOURTH_FLOOR", _("Fourth Floor")
    FIFTH_FLOOR = "FIFTH_FLOOR", _("Fifth Floor")
    SIXTH_FLOOR_PLUS = "SIXTH_FLOOR_PLUS", _("Sixth Floor Plus")


class LocationChoicesEnum(models.TextChoices):
    HOME = "HOME", _("Home")
    OFFICE = "OFFICE", _("Office")
    OTHER = "OTHER", _("Other")



from django.contrib import admin
from django.contrib.auth.admin import GroupAdmin as BaseGroupAdmin
from django.contrib.auth.models import Group
from django.db import models
from django.db.models import Count, Q
from django.utils import timezone
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from parler.admin import TranslatableAdmin
from unfold.admin import ModelAdmin, TabularInline
from unfold.contrib.filters.admin import (
    DropdownFilter,
    RangeDateFilter,
    RangeDateTimeFilter,
    RangeNumericListFilter,
    RelatedDropdownFilter,
)
from unfold.contrib.forms.widgets import WysiwygWidget
from unfold.decorators import action
from unfold.enums import ActionVariant
from unfold.forms import (
    AdminPasswordChangeForm,
    UserChangeForm,
    UserCreationForm,
)

from user.models import UserAccount
from user.models.address import UserAddress
from user.models.subscription import SubscriptionTopic, UserSubscription

admin.site.unregister(Group)


class SubscriptionCountFilter(RangeNumericListFilter):
    title = _("Subscription Count")
    parameter_name = "subscription_count"

    def queryset(self, request, queryset):
        filters = {}

        value_from = self.used_parameters.get(f"{self.parameter_name}_from")
        if value_from:
            filters["subscription_count__gte"] = value_from

        value_to = self.used_parameters.get(f"{self.parameter_name}_to")
        if value_to:
            filters["subscription_count__lte"] = value_to

        return queryset.filter(**filters) if filters else queryset

    def expected_parameters(self):
        return [f"{self.parameter_name}_from", f"{self.parameter_name}_to"]


class AddressCountFilter(RangeNumericListFilter):
    title = _("Address Count")
    parameter_name = "address_count"

    def queryset(self, request, queryset):
        filters = {}

        value_from = self.used_parameters.get(f"{self.parameter_name}_from")
        if value_from:
            filters["address_count__gte"] = value_from

        value_to = self.used_parameters.get(f"{self.parameter_name}_to")
        if value_to:
            filters["address_count__lte"] = value_to

        return queryset.filter(**filters) if filters else queryset

    def expected_parameters(self):
        return [f"{self.parameter_name}_from", f"{self.parameter_name}_to"]


class UserStatusFilter(DropdownFilter):
    title = _("User Status")
    parameter_name = "user_status"

    def lookups(self, request, model_admin):
        return [
            ("active_staff", _("Active Staff")),
            ("active_regular", _("Active Regular Users")),
            ("inactive", _("Inactive Users")),
            ("superuser", _("Super Users")),
            ("with_subscriptions", _("Users with Subscriptions")),
            ("no_subscriptions", _("Users without Subscriptions")),
        ]

    def queryset(self, request, queryset):
        v = self.value()
        if v == "active_staff":
            return queryset.filter(is_active=True, is_staff=True)
        if v == "active_regular":
            return queryset.filter(
                is_active=True, is_staff=False, is_superuser=False
            )
        if v == "inactive":
            return queryset.filter(is_active=False)
        if v == "superuser":
            return queryset.filter(is_superuser=True)
        if v == "with_subscriptions":
            return queryset.filter(subscriptions__isnull=False).distinct()
        if v == "no_subscriptions":
            return queryset.filter(subscriptions__isnull=True)
        return queryset


class SocialMediaFilter(DropdownFilter):
    title = _("Social Media Presence")
    parameter_name = "social_media"

    def lookups(self, request, model_admin):
        return [
            ("has_website", _("Has Website")),
            ("has_social", _("Has Social Media")),
            ("no_social", _("No Social Media")),
        ]

    def queryset(self, request, queryset):
        v = self.value()
        if v == "has_website":
            return queryset.exclude(website="")
        if v == "has_social":
            return queryset.filter(
                Q(twitter__gt="")
                | Q(linkedin__gt="")
                | Q(facebook__gt="")
                | Q(instagram__gt="")
                | Q(youtube__gt="")
                | Q(github__gt="")
            )
        if v == "no_social":
            return queryset.filter(
                website="",
                twitter="",
                linkedin="",
                facebook="",
                instagram="",
                youtube="",
                github="",
            )
        return queryset


class TopicCategoryFilter(DropdownFilter):
    title = _("Topic Category")
    parameter_name = "topic_category"

    def lookups(self, request, model_admin):
        return SubscriptionTopic.TopicCategory.choices

    def queryset(self, request, queryset):
        v = self.value()
        return queryset.filter(topic__category=v) if v else queryset


class UserAddressInline(TabularInline):
    model = UserAddress
    extra = 0
    fields = (
        "title",
        "first_name",
        "last_name",
        "city",
        "country",
        "is_main",
        "phone",
    )
    readonly_fields = ("created_at",)
    show_change_link = True
    tab = True


class UserSubscriptionInline(TabularInline):
    model = UserSubscription
    extra = 0
    fields = ("topic", "status", "subscribed_at", "unsubscribed_at")
    readonly_fields = ("subscribed_at", "unsubscribed_at")
    show_change_link = True
    tab = True


@admin.register(Group)
class GroupAdmin(BaseGroupAdmin, ModelAdmin):
    pass


@admin.register(UserAccount)
class UserAdmin(ModelAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = True
    list_filter_submit = True
    list_filter_sheet = True

    form = UserChangeForm
    add_form = UserCreationForm
    change_password_form = AdminPasswordChangeForm

    formfield_overrides = {models.TextField: {"widget": WysiwygWidget}}

    list_display = [
        "user_profile_display",
        "contact_info_display",
        "location_display",
        "user_status_badges",
        "social_links_display",
        "engagement_metrics",
        "last_activity",
        "created_at",
    ]

    list_filter = [
        UserStatusFilter,
        SocialMediaFilter,
        "is_active",
        "is_staff",
        "is_superuser",
        ("country", RelatedDropdownFilter),
        ("region", RelatedDropdownFilter),
        SubscriptionCountFilter,
        AddressCountFilter,
        ("created_at", RangeDateTimeFilter),
        ("updated_at", RangeDateTimeFilter),
        ("birth_date", RangeDateFilter),
    ]

    search_fields = [
        "email",
        "username",
        "first_name",
        "last_name",
        "phone",
        "city",
        "address",
        "bio",
    ]

    list_select_related = ["country", "region"]
    readonly_fields = [
        "id",
        "created_at",
        "updated_at",
        "engagement_metrics",
        "social_links_summary",
        "subscription_summary",
        "address_summary",
    ]

    ordering = ["-created_at"]
    date_hierarchy = "created_at"
    save_on_top = True
    list_per_page = 25

    fieldsets = (
        (
            _("Account Credentials"),
            {"fields": ("email", "username", "password"), "classes": ("wide",)},
        ),
        (
            _("Personal Information"),
            {
                "fields": (
                    "first_name",
                    "last_name",
                    "phone",
                    "birth_date",
                    "image",
                ),
                "classes": ("wide",),
            },
        ),
        (_("Bio & Description"), {"fields": ("bio",), "classes": ("wide",)}),
        (
            _("Location & Address"),
            {
                "fields": (
                    "address",
                    "city",
                    "zipcode",
                    "place",
                    "country",
                    "region",
                    "address_summary",
                ),
                "classes": ("wide",),
            },
        ),
        (
            _("Social Media & Website"),
            {
                "fields": (
                    "website",
                    "linkedin",
                    "github",
                    "twitter",
                    "facebook",
                    "instagram",
                    "youtube",
                    "social_links_summary",
                ),
                "classes": ("wide",),
            },
        ),
        (
            _("Account Permissions"),
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                ),
                "classes": ("wide",),
            },
        ),
        (
            _("Subscriptions & Engagement"),
            {
                "fields": ("subscription_summary", "engagement_metrics"),
                "classes": ("collapse",),
            },
        ),
        (
            _("System Information"),
            {
                "fields": ("id", "created_at", "updated_at"),
                "classes": ("collapse",),
            },
        ),
    )

    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("email", "username", "password1", "password2"),
            },
        ),
    )
    inlines = [UserAddressInline, UserSubscriptionInline]

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.annotate(
            subscription_count=Count("subscriptions", distinct=True),
            address_count=Count("addresses", distinct=True),
            active_subscription_count=Count(
                "subscriptions",
                filter=Q(
                    subscriptions__status=UserSubscription.SubscriptionStatus.ACTIVE
                ),
                distinct=True,
            ),
        )

    def user_profile_display(self, obj):
        html = '<div class="flex items-center gap-3">'
        if obj.image:
            url = conditional_escape(obj.image.url)
            html += f'<img src="{url}" class="rounded-full object-cover" style="width:60px;height:60px;object-fit:cover;border-radius:8px;border:1px solid #e5e7eb" />'
        else:
            initials = "".join(
                [n[0].upper() for n in [obj.first_name, obj.last_name] if n]
            )[:2] or (obj.email[0].upper() if obj.email else "U")
            esc_init = conditional_escape(initials)
            html += f'<div class="rounded-full bg-blue-100 dark:bg-blue-900 flex items-center justify-center text-blue-600 dark:text-blue-300 font-medium text-sm" style="width:60px;height:60px;object-fit:cover;border-radius:8px;border:1px solid #e5e7eb">{esc_init}</div>'
        full_name = obj.full_name or obj.username or "Anonymous"
        esc_name = conditional_escape(full_name)
        esc_email = conditional_escape(obj.email or "")
        html += f'<div><div class="font-medium text-base-900 dark:text-base-100">{esc_name}</div>'
        html += f'<div class="text-sm text-base-500 dark:text-base-400">{esc_email}</div></div></div>'
        return mark_safe(html)

    user_profile_display.short_description = _("Profile")

    def contact_info_display(self, obj):
        parts = []
        if obj.phone:
            esc_phone = conditional_escape(obj.phone)
            parts.append(
                f'<span class="flex items-center gap-1"><span>📞</span><span>{esc_phone}</span></span>'
            )
        if obj.username:
            esc_user = conditional_escape(obj.username)
            parts.append(
                f'<span class="flex items-center gap-1"><span>👤</span><span>{esc_user}</span></span>'
            )
        if parts:
            html = (
                '<div class="text-sm text-base-700 dark:text-base-300 space-y-1">'
                + "".join(parts)
                + "</div>"
            )
            return mark_safe(html)
        return mark_safe(
            '<span class="text-base-400 dark:text-base-500">No contact info</span>'
        )

    contact_info_display.short_description = _("Contact")

    def location_display(self, obj):
        loc = []
        if obj.city:
            loc.append(conditional_escape(obj.city))
        if obj.country:
            loc.append(conditional_escape(str(obj.country)))
        if loc:
            text = ", ".join(loc)
            html = f'<div class="text-sm text-base-700 dark:text-base-300"><span class="flex items-center gap-1"><span>📍</span><span>{text}</span></span></div>'
            return mark_safe(html)
        return mark_safe(
            '<span class="text-base-400 dark:text-base-500">No location</span>'
        )

    location_display.short_description = _("Location")

    def user_status_badges(self, obj):
        badges = []
        if obj.is_superuser:
            badges.append(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300 rounded-full gap-1"><span>👑</span><span>Super</span></span>'
            )
        elif obj.is_staff:
            badges.append(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-purple-50 dark:bg-purple-900 text-purple-700 dark:text-purple-300 rounded-full gap-1"><span>⚡</span><span>Staff</span></span>'
            )
        if obj.is_active:
            badges.append(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-300 rounded-full gap-1"><span>✓</span><span>Active</span></span>'
            )
        else:
            badges.append(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300 rounded-full gap-1"><span>✗</span><span>Inactive</span></span>'
            )
        return mark_safe(
            '<div class="flex flex-wrap gap-1">' + "".join(badges) + "</div>"
        )

    user_status_badges.short_description = _("Status")

    def social_links_display(self, obj):
        icons = []
        for field, (icon, icon_name) in {
            "website": ("🌐", "Website"),
            "linkedin": ("💼", "LinkedIn"),
            "github": ("💻", "GitHub"),
            "twitter": ("🐦", "Twitter"),
        }.items():
            val = getattr(obj, field)
            if val:
                icons.append(conditional_escape(icon))
        if icons:
            return mark_safe(
                '<div class="flex gap-1">' + "".join(icons) + "</div>"
            )
        return mark_safe(
            '<span class="text-base-400 dark:text-base-500">-</span>'
        )

    social_links_display.short_description = _("Social")

    def engagement_metrics(self, obj):
        subs = getattr(obj, "subscription_count", 0)
        active = getattr(obj, "active_subscription_count", 0)
        addrs = getattr(obj, "address_count", 0)
        esc_active = conditional_escape(str(active))
        esc_subs = conditional_escape(str(subs))
        esc_addrs = conditional_escape(str(addrs))
        html = (
            '<div class="text-sm text-base-700 dark:text-base-300 flex items-center gap-3">'
            f'<span class="flex items-center gap-1 text-blue-600 dark:text-blue-400"><span>📧</span><span>{esc_active}/{esc_subs}</span></span>'
            f'<span class="flex items-center gap-1 text-green-600 dark:text-green-400"><span>📍</span><span>{esc_addrs}</span></span>'
            "</div>"
        )
        return mark_safe(html)

    engagement_metrics.short_description = _("Engagement")

    def last_activity(self, obj):
        ts = (
            obj.updated_at.strftime("%Y-%m-%d %H:%M")
            if obj.updated_at
            else "Never"
        )
        esc_ts = conditional_escape(ts)
        html = f'<div class="text-sm text-base-600 dark:text-base-400"><div>Updated: {esc_ts}</div></div>'
        return mark_safe(html)

    last_activity.short_description = _("Last Activity")

    def social_links_summary(self, obj):
        links = []
        fields = {
            "website": "Website",
            "linkedin": "LinkedIn",
            "github": "GitHub",
            "twitter": "Twitter",
            "facebook": "Facebook",
            "instagram": "Instagram",
            "youtube": "YouTube",
        }
        for fld, name in fields.items():
            url = getattr(obj, fld)
            if url:
                esc_url = conditional_escape(url)
                links.append(
                    f'<a href="{esc_url}" target="_blank" class="text-blue-600 dark:text-blue-400 hover:underline">{conditional_escape(name)}</a>'
                )
        if links:
            return mark_safe(
                '<div class="space-y-1">' + "<br>".join(links) + "</div>"
            )
        return mark_safe(
            '<span class="text-base-400 dark:text-base-500 italic">No social media links</span>'
        )

    social_links_summary.short_description = _("Social Links Summary")

    def subscription_summary(self, obj):
        subs = list(obj.subscriptions.select_related("topic").all())
        if not subs:
            return mark_safe(
                '<span class="text-base-400 dark:text-base-500 italic">No subscriptions</span>'
            )
        active = sum(
            1
            for s in subs
            if s.status == UserSubscription.SubscriptionStatus.ACTIVE
        )
        total = len(subs)
        esc_active = conditional_escape(str(active))
        esc_total = conditional_escape(str(total))
        html = f'<div class="text-sm"><div class="font-medium text-base-700 dark:text-base-300">Active: {esc_active}/{esc_total}</div></div>'
        return mark_safe(html)

    subscription_summary.short_description = _("Subscription Summary")

    def address_summary(self, obj):
        addrs = list(obj.addresses.all())
        if not addrs:
            return mark_safe(
                '<span class="text-base-400 dark:text-base-500 italic">No addresses</span>'
            )
        main = next((a for a in addrs if a.is_main), None)
        esc_total = conditional_escape(str(len(addrs)))
        main_text = conditional_escape(str(main)) if main else "No main address"
        html = f'<div class="text-sm"><div class="font-medium text-base-700 dark:text-base-300">Total: {esc_total}</div><div class="text-base-500 dark:text-base-400">Main: {main_text}</div></div>'
        return mark_safe(html)

    address_summary.short_description = _("Address Summary")


@admin.register(UserAddress)
class UserAddressAdmin(ModelAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = True
    list_filter_submit = True
    list_filter_sheet = True

    list_display = [
        "address_display",
        "contact_person",
        "location_info",
        "main_address_badge",
        "contact_numbers",
        "created_at",
    ]

    list_filter = [
        "is_main",
        "floor",
        "location_type",
        ("country", RelatedDropdownFilter),
        ("region", RelatedDropdownFilter),
        ("created_at", RangeDateTimeFilter),
    ]

    search_fields = [
        "user__email",
        "user__username",
        "title",
        "first_name",
        "last_name",
        "street",
        "city",
    ]

    list_select_related = ["user", "country", "region"]
    readonly_fields = ["id", "created_at", "updated_at"]

    fieldsets = (
        (
            _("Address Owner"),
            {"fields": ("user", "title"), "classes": ("wide",)},
        ),
        (
            _("Contact Person"),
            {"fields": ("first_name", "last_name"), "classes": ("wide",)},
        ),
        (
            _("Address Details"),
            {
                "fields": (
                    "street",
                    "street_number",
                    "city",
                    "zipcode",
                    "country",
                    "region",
                ),
                "classes": ("wide",),
            },
        ),
        (
            _("Additional Information"),
            {
                "fields": ("floor", "location_type", "notes", "is_main"),
                "classes": ("wide",),
            },
        ),
        (
            _("Contact Information"),
            {"fields": ("phone", "mobile_phone"), "classes": ("wide",)},
        ),
        (
            _("System Information"),
            {
                "fields": ("id", "created_at", "updated_at"),
                "classes": ("collapse",),
            },
        ),
    )

    def address_display(self, obj):
        esc_title = conditional_escape(obj.title)
        addr = f"{obj.street} {obj.street_number}, {obj.city}"
        esc_addr = conditional_escape(addr)
        html = f'<div class="text-sm"><div class="font-medium text-base-900 dark:text-base-100">{esc_title}</div><div class="text-base-500 dark:text-base-400">{esc_addr}</div></div>'
        return mark_safe(html)

    address_display.short_description = _("Address")

    def contact_person(self, obj):
        esc_name = conditional_escape(f"{obj.first_name} {obj.last_name}")
        esc_email = conditional_escape(obj.user.email)
        html = f'<div class="text-sm"><div class="font-medium text-base-700 dark:text-base-300">{esc_name}</div><div class="text-base-500 dark:text-base-400">{esc_email}</div></div>'
        return mark_safe(html)

    contact_person.short_description = _("Contact Person")

    def location_info(self, obj):
        parts = [conditional_escape(obj.city)]
        if obj.country:
            parts.append(conditional_escape(str(obj.country)))
        if obj.zipcode:
            parts.append(conditional_escape(obj.zipcode))
        text = ", ".join(parts)
        html = f'<div class="text-sm text-base-700 dark:text-base-300">{text}</div>'
        return mark_safe(html)

    location_info.short_description = _("Location")

    def main_address_badge(self, obj):
        if obj.is_main:
            return mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-300 rounded-full gap-1"><span>⭐</span><span>Main</span></span>'
            )
        return ""

    main_address_badge.short_description = _("Main")

    def contact_numbers(self, obj):
        nums = []
        if obj.phone:
            esc_phone = conditional_escape(obj.phone)
            nums.append(
                f'<span class="flex items-center gap-1"><span>📞</span><span>{esc_phone}</span></span>'
            )
        if obj.mobile_phone:
            esc_mobile = conditional_escape(obj.mobile_phone)
            nums.append(
                f'<span class="flex items-center gap-1"><span>📱</span><span>{esc_mobile}</span></span>'
            )
        if nums:
            html = (
                '<div class="text-sm text-base-700 dark:text-base-300 space-y-1">'
                + "".join(nums)
                + "</div>"
            )
            return mark_safe(html)
        return mark_safe(
            '<span class="text-base-400 dark:text-base-500">No phone</span>'
        )

    contact_numbers.short_description = _("Contact Numbers")


@admin.register(SubscriptionTopic)
class SubscriptionTopicAdmin(ModelAdmin, TranslatableAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = False
    list_filter_submit = True
    list_filter_sheet = True

    list_display = [
        "name_display",
        "category_badge",
        "active_status",
        "is_active",
        "settings_badges",
        "subscriber_metrics",
        "created_at",
    ]
    list_filter = [
        "category",
        "is_active",
        "is_default",
        "requires_confirmation",
        ("created_at", RangeDateTimeFilter),
    ]
    search_fields = ["translations__name", "slug", "translations__description"]
    readonly_fields = ["uuid", "created_at", "updated_at", "subscriber_metrics"]

    fieldsets = (
        (
            _("Topic Information"),
            {
                "fields": ("name", "slug", "description", "category"),
                "classes": ("wide",),
            },
        ),
        (
            _("Settings"),
            {
                "fields": ("is_active", "is_default", "requires_confirmation"),
                "classes": ("wide",),
            },
        ),
        (
            _("System Information"),
            {
                "fields": (
                    "uuid",
                    "created_at",
                    "updated_at",
                    "subscriber_metrics",
                ),
                "classes": ("collapse",),
            },
        ),
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.annotate(
            active_subscribers=Count(
                "subscribers",
                filter=Q(
                    subscribers__status=UserSubscription.SubscriptionStatus.ACTIVE
                ),
            ),
            total_subscribers=Count("subscribers"),
        )

    def name_display(self, obj):
        name = (
            obj.safe_translation_getter("name", any_language=True)
            or "Unnamed Topic"
        )
        esc_name = conditional_escape(name)
        esc_slug = conditional_escape(obj.slug)
        html = f'<div class="text-sm"><div class="font-medium text-base-900 dark:text-base-100">{esc_name}</div><div class="text-base-500 dark:text-base-400">{esc_slug}</div></div>'
        return mark_safe(html)

    name_display.short_description = _("Topic")

    def category_badge(self, obj):
        colors = {
            "MARKETING": "bg-purple-50 dark:bg-purple-900 text-purple-700 dark:text-purple-300",
            "PRODUCT": "bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-300",
            "ACCOUNT": "bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-300",
            "SYSTEM": "bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300",
            "NEWSLETTER": "bg-indigo-50 dark:bg-indigo-900 text-indigo-700 dark:text-indigo-300",
            "OTHER": "bg-base-50 dark:bg-base-900 text-base-700 dark:text-base-300",
        }
        cls = conditional_escape(colors.get(obj.category, colors["OTHER"]))
        disp = conditional_escape(obj.get_category_display())
        html = f'<span class="inline-flex items-center px-2 py-1 text-xs font-medium {cls} rounded-full">{disp}</span>'
        return mark_safe(html)

    category_badge.short_description = _("Category")

    def active_status(self, obj):
        if obj.is_active:
            html = '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-300 rounded-full gap-1"><span>✓</span><span>Active</span></span>'
        else:
            html = '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300 rounded-full gap-1"><span>✗</span><span>Inactive</span></span>'
        return mark_safe(html)

    active_status.short_description = _("Status")

    def settings_badges(self, obj):
        badges = []
        if obj.is_default:
            badges.append(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-300 rounded-full gap-1"><span>⭐</span><span>Default</span></span>'
            )
        if obj.requires_confirmation:
            badges.append(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-orange-50 dark:bg-orange-900 text-orange-700 dark:text-orange-300 rounded-full gap-1"><span>✉️</span><span>Confirm</span></span>'
            )
        return (
            mark_safe(
                '<div class="flex flex-wrap gap-1">'
                + "".join(badges)
                + "</div>"
            )
            if badges
            else ""
        )

    settings_badges.short_description = _("Settings")

    def subscriber_metrics(self, obj):
        active = conditional_escape(str(getattr(obj, "active_subscribers", 0)))
        total = conditional_escape(str(getattr(obj, "total_subscribers", 0)))
        html = f'<div class="text-sm text-base-700 dark:text-base-300 flex items-center gap-3"><span class="flex items-center gap-1 text-green-600 dark:text-green-400"><span>✓</span><span>{active}</span></span><span class="flex items-center gap-1 text-blue-600 dark:text-blue-400"><span>👥</span><span>{total}</span></span></div>'
        return mark_safe(html)

    subscriber_metrics.short_description = _("Subscribers")


@admin.register(UserSubscription)
class UserSubscriptionAdmin(ModelAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = True
    list_filter_submit = True
    list_filter_sheet = True

    list_display = [
        "subscription_info",
        "user_info",
        "topic_info",
        "status_display",
        "status",
        "subscription_dates",
        "created_at",
    ]

    list_filter = [
        "status",
        TopicCategoryFilter,
        ("topic", RelatedDropdownFilter),
        ("subscribed_at", RangeDateTimeFilter),
        ("unsubscribed_at", RangeDateTimeFilter),
    ]
    search_fields = [
        "user__email",
        "user__username",
        "topic__translations__name",
        "user__first_name",
        "user__last_name",
    ]
    readonly_fields = [
        "subscribed_at",
        "unsubscribed_at",
        "created_at",
        "updated_at",
    ]
    raw_id_fields = ["user"]
    autocomplete_fields = ["topic"]
    list_select_related = ["user", "topic"]

    actions = ["activate_subscriptions", "deactivate_subscriptions"]

    fieldsets = (
        (
            _("Subscription Details"),
            {"fields": ("user", "topic", "status"), "classes": ("wide",)},
        ),
        (
            _("Timestamps"),
            {
                "fields": (
                    "subscribed_at",
                    "unsubscribed_at",
                    "created_at",
                    "updated_at",
                ),
                "classes": ("wide",),
            },
        ),
        (
            _("Additional Data"),
            {
                "fields": ("confirmation_token", "metadata"),
                "classes": ("collapse",),
            },
        ),
    )

    def subscription_info(self, obj):
        esc_id = conditional_escape(str(obj.id))
        date = conditional_escape(obj.created_at.strftime("%Y-%m-%d"))
        html = f'<div class="text-sm"><div class="font-medium text-base-900 dark:text-base-100">Subscription #{esc_id}</div><div class="text-base-500 dark:text-base-400">{date}</div></div>'
        return mark_safe(html)

    subscription_info.short_description = _("Subscription")

    def user_info(self, obj):
        name = obj.user.full_name or obj.user.username or "Anonymous"
        esc_name = conditional_escape(name)
        esc_email = conditional_escape(obj.user.email)
        html = f'<div class="text-sm"><div class="font-medium text-base-700 dark:text-base-300">{esc_name}</div><div class="text-base-500 dark:text-base-400">{esc_email}</div></div>'
        return mark_safe(html)

    user_info.short_description = _("User")

    def topic_info(self, obj):
        topic = (
            obj.topic.safe_translation_getter("name", any_language=True)
            or "Unnamed Topic"
        )
        esc_topic = conditional_escape(topic)
        esc_cat = conditional_escape(obj.topic.get_category_display())
        html = f'<div class="text-sm"><div class="font-medium text-base-700 dark:text-base-300">{esc_topic}</div><div class="text-base-500 dark:text-base-400">{esc_cat}</div></div>'
        return mark_safe(html)

    topic_info.short_description = _("Topic")

    def status_display(self, obj):
        colors = {
            UserSubscription.SubscriptionStatus.ACTIVE: "bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-300",
            UserSubscription.SubscriptionStatus.PENDING: "bg-orange-50 dark:bg-orange-900 text-orange-700 dark:text-orange-300",
            UserSubscription.SubscriptionStatus.UNSUBSCRIBED: "bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300",
            UserSubscription.SubscriptionStatus.BOUNCED: "bg-purple-50 dark:bg-purple-900 text-purple-700 dark:text-purple-300",
        }
        icons = {
            UserSubscription.SubscriptionStatus.ACTIVE: "✓",
            UserSubscription.SubscriptionStatus.PENDING: "⏳",
            UserSubscription.SubscriptionStatus.UNSUBSCRIBED: "✗",
            UserSubscription.SubscriptionStatus.BOUNCED: "⚠️",
        }
        cls = conditional_escape(
            colors.get(
                obj.status, colors[UserSubscription.SubscriptionStatus.ACTIVE]
            )
        )
        icon = conditional_escape(icons.get(obj.status, "?"))
        label = conditional_escape(obj.get_status_display())
        html = f'<span class="inline-flex items-center px-2 py-1 text-xs font-medium {cls} rounded-full gap-1"><span>{icon}</span><span>{label}</span></span>'
        return mark_safe(html)

    status_display.short_description = _("Status")

    def subscription_dates(self, obj):
        sub = conditional_escape(obj.subscribed_at.strftime("%Y-%m-%d %H:%M"))
        html = f'<div class="text-sm text-base-600 dark:text-base-400"><div>Subscribed: {sub}</div>'
        if obj.unsubscribed_at:
            unsub = conditional_escape(
                obj.unsubscribed_at.strftime("%Y-%m-%d %H:%M")
            )
            html += f"<div>Unsubscribed: {unsub}</div>"
        html += "</div>"
        return mark_safe(html)

    subscription_dates.short_description = _("Dates")

    @action(
        description=_("Activate selected subscriptions"),
        variant=ActionVariant.SUCCESS,
        icon="check_circle",
    )
    def activate_subscriptions(self, request, queryset):
        updated = queryset.filter(
            status__in=[
                UserSubscription.SubscriptionStatus.PENDING,
                UserSubscription.SubscriptionStatus.UNSUBSCRIBED,
            ]
        ).update(
            status=UserSubscription.SubscriptionStatus.ACTIVE,
            unsubscribed_at=None,
        )
        self.message_user(
            request,
            _("%(count)d subscriptions were activated.") % {"count": updated},
        )

    @action(
        description=_("Deactivate selected subscriptions"),
        variant=ActionVariant.WARNING,
        icon="cancel",
    )
    def deactivate_subscriptions(self, request, queryset):
        updated = queryset.filter(
            status=UserSubscription.SubscriptionStatus.ACTIVE
        ).update(
            status=UserSubscription.SubscriptionStatus.UNSUBSCRIBED,
            unsubscribed_at=timezone.now(),
        )
        self.message_user(
            request,
            _("%(count)d subscriptions were deactivated.") % {"count": updated},
        )



from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema_field
from parler_rest.serializers import TranslatableModelSerializer
from rest_framework import serializers

from user.serializers.account import UserDetailsSerializer
from core.api.schema import generate_schema_multi_lang
from core.utils.serializers import TranslatedFieldExtended
from product.models.product import Product
from product.models.review import ProductReview
from product.serializers.product import ProductSerializer

User = get_user_model()


@extend_schema_field(generate_schema_multi_lang(ProductReview))
class TranslatedFieldsFieldExtend(TranslatedFieldExtended):
    pass


class ProductReviewSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[ProductReview]
):
    translations = TranslatedFieldsFieldExtend(shared_model=ProductReview)
    user = UserDetailsSerializer(read_only=True)
    product = ProductSerializer(read_only=True)

    class Meta:
        model = ProductReview
        fields = (
            "id",
            "product",
            "user",
            "rate",
            "status",
            "is_published",
            "created_at",
            "updated_at",
            "published_at",
            "uuid",
            "translations",
        )
        read_only_fields = (
            "id",
            "created_at",
            "updated_at",
            "published_at",
            "uuid",
        )


class ProductReviewDetailSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[ProductReview]
):
    translations = TranslatedFieldsFieldExtend(shared_model=ProductReview)
    user = UserDetailsSerializer(read_only=True)
    product = ProductSerializer(read_only=True)

    class Meta:
        model = ProductReview
        fields = (
            "id",
            "product",
            "user",
            "rate",
            "status",
            "is_published",
            "created_at",
            "updated_at",
            "published_at",
            "uuid",
            "translations",
        )
        read_only_fields = (
            "id",
            "created_at",
            "updated_at",
            "published_at",
            "uuid",
        )


class ProductReviewWriteSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[ProductReview]
):
    translations = TranslatedFieldsFieldExtend(shared_model=ProductReview)
    product = serializers.PrimaryKeyRelatedField(queryset=Product.objects.all())

    class Meta:
        model = ProductReview
        fields = (
            "product",
            "rate",
            "status",
            "is_published",
            "translations",
        )

    def validate_rate(self, value: int) -> int:
        valid_rates = [choice[0] for choice in ProductReview.rate.field.choices]
        if value not in valid_rates:
            raise serializers.ValidationError(_("Invalid rate value."))
        return value

    def create(self, validated_data):
        user = self.context["request"].user
        validated_data["user"] = user
        return super().create(validated_data)

    def validate(self, attrs):
        user = self.context["request"].user
        product = attrs.get("product")

        if (
            self.instance is None
            and ProductReview.objects.filter(
                user=user, product=product
            ).exists()
        ):
            raise serializers.ValidationError(
                _("You have already reviewed this product.")
            )

        return attrs



import uuid

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory, TestCase

from cart.factories.cart import CartFactory
from cart.factories.item import CartItemFactory
from cart.models import Cart, CartItem
from cart.services import (
    CartService,
)
from product.factories.product import ProductFactory
from user.factories.account import UserAccountFactory

User = get_user_model()


class CartServiceTest(TestCase):
    def setUp(self):
        self.user = UserAccountFactory(num_addresses=0)
        self.factory = RequestFactory()
        self.product = ProductFactory(num_images=0, num_reviews=0)

        self.cart = CartFactory(user=self.user, num_cart_items=0)
        CartItem.objects.all().delete()

        self.request = self._create_request_with_headers(
            user=self.user,
            cart_id=self.cart.id,
            session_key=None,
        )

    def _create_request_with_headers(
        self, user=None, cart_id=None, session_key=None
    ):
        request = self.factory.get("/")
        request.user = user or AnonymousUser()

        if cart_id:
            request.META["HTTP_X_CART_ID"] = str(cart_id)
        if session_key:
            request.META["HTTP_X_SESSION_KEY"] = session_key

        request.session = {}

        return request

    def test_get_cart_item(self):
        cart_service = CartService(request=self.request)
        cart_service.create_cart_item(self.product, 2)
        cart_item = cart_service.get_cart_item(self.product.pk)
        self.assertEqual(cart_item.product, self.product)

    def test_create_cart_item(self):
        CartItem.objects.all().delete()
        cart_service = CartService(request=self.request)
        cart_item = cart_service.create_cart_item(self.product, 3)
        self.assertEqual(cart_item.quantity, 3)
        self.assertEqual(CartItem.objects.count(), 1)

    def test_update_cart_item(self):
        cart_service = CartService(request=self.request)
        cart_service.create_cart_item(self.product, 2)
        cart_item = cart_service.update_cart_item(self.product.pk, 5)
        self.assertEqual(cart_item.quantity, 5)

    def test_delete_cart_item(self):
        cart_service = CartService(request=self.request)
        cart_service.create_cart_item(self.product, 2)
        cart_service.delete_cart_item(self.product.pk)
        self.assertEqual(CartItem.objects.count(), 0)

    def test_process_cart_merge_with_guest_cart_headers(self):
        guest_cart = CartFactory(user=None, num_cart_items=0)
        guest_product = ProductFactory(num_images=0, num_reviews=0)
        CartItemFactory(cart=guest_cart, product=guest_product, quantity=2)

        request = self._create_request_with_headers(
            user=self.user,
            cart_id=guest_cart.id,
            session_key=guest_cart.session_key,
        )

        CartService(request=request)

        self.assertEqual(self.cart.items.count(), 1)
        self.assertEqual(self.cart.items.first().quantity, 2)
        self.assertEqual(self.cart.items.first().product, guest_product)

        self.assertFalse(Cart.objects.filter(id=guest_cart.id).exists())

    def test_get_cart_by_id_valid(self):
        cart_service = CartService(request=self.request)
        cart_id = self.cart.id

        cart = cart_service.get_cart_by_id(cart_id)

        self.assertEqual(cart, self.cart)

    def test_get_cart_by_id_invalid(self):
        cart_service = CartService(request=self.request)
        cart_id = self.cart.id + 1000

        cart = cart_service.get_cart_by_id(cart_id)

        self.assertIsNone(cart)

    def test_cart_service_string_representation(self):
        cart_service = CartService(request=self.request)
        self.assertEqual(str(cart_service), f"Cart {self.user}")

    def test_cart_service_length(self):
        cart_service = CartService(request=self.request)
        cart_service.create_cart_item(self.product, 3)

        self.assertEqual(len(cart_service), 3)

    def test_cart_service_iteration(self):
        cart_service = CartService(request=self.request)
        cart_service.create_cart_item(self.product, 2)

        cart_items = list(cart_service)

        self.assertEqual(len(cart_items), 1)
        self.assertEqual(cart_items[0].product, self.product)

    def test_cart_service_add_more_items(self):
        cart_service = CartService(request=self.request)

        product = ProductFactory(num_images=0, num_reviews=0)

        cart_service.create_cart_item(self.product, 3)
        cart_service.create_cart_item(product, 4)

        self.assertEqual(len(cart_service), 7)
        self.assertEqual(self.cart.items.count(), 2)

    def test_cart_service_remove_item(self):
        cart_service = CartService(request=self.request)
        cart_service.create_cart_item(self.product, 2)

        cart_service.delete_cart_item(self.product.pk)

        self.assertEqual(len(cart_service), 0)
        self.assertEqual(self.cart.items.count(), 0)

    def test_merge_carts_with_items_in_both_carts(self):
        cart_service = CartService(request=self.request)
        cart_service.create_cart_item(self.product, 2)

        guest_cart = CartFactory(user=None, num_cart_items=0)
        guest_product = ProductFactory(num_images=0, num_reviews=0)
        CartItemFactory(cart=guest_cart, product=guest_product, quantity=1)

        cart_service.merge_carts(guest_cart)

        self.assertEqual(self.cart.items.count(), 2)
        self.assertCountEqual(
            [item.product for item in self.cart.items.all()],
            [self.product, guest_product],
        )


class GuestCartServiceTest(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.session_key = str(uuid.uuid4())
        self.product = ProductFactory(num_images=0, num_reviews=0)

    def _create_guest_request(self, cart_id=None, session_key=None):
        request = self.factory.get("/")
        request.user = AnonymousUser()

        if cart_id:
            request.META["HTTP_X_CART_ID"] = str(cart_id)
        if session_key:
            request.META["HTTP_X_SESSION_KEY"] = session_key

        request.session = {}
        return request

    def test_get_or_create_cart_for_new_guest(self):
        request = self._create_guest_request(session_key=self.session_key)

        cart_service = CartService(request=request)
        cart = cart_service.get_or_create_cart()

        self.assertIsNotNone(cart)
        self.assertIsNone(cart.user)
        self.assertEqual(cart.session_key, self.session_key)

    def test_get_or_create_cart_for_guest_with_existing_cart(self):
        guest_cart = CartFactory(user=None, session_key=self.session_key)

        request = self._create_guest_request(
            cart_id=guest_cart.id, session_key=self.session_key
        )

        cart_service = CartService(request=request)
        cart = cart_service.get_or_create_cart()

        self.assertIsNotNone(cart)
        self.assertEqual(cart.id, guest_cart.id)
        self.assertEqual(cart.session_key, self.session_key)

    def test_create_cart_item_for_guest(self):
        guest_cart = CartFactory(user=None, session_key=self.session_key)
        request = self._create_guest_request(
            cart_id=guest_cart.id, session_key=self.session_key
        )

        cart_service = CartService(request=request)
        cart_item = cart_service.create_cart_item(self.product, 3)

        self.assertEqual(cart_item.quantity, 3)
        self.assertEqual(cart_item.product, self.product)
        self.assertIsNone(cart_item.cart.user)
        self.assertEqual(cart_item.cart.id, guest_cart.id)

    def test_guest_cart_string_representation(self):
        request = self._create_guest_request(session_key=self.session_key)
        cart_service = CartService(request=request)

        self.assertTrue(str(cart_service).startswith("Cart Anonymous"))

    def test_guest_cart_with_mismatched_session_key(self):
        existing_cart = CartFactory(user=None, session_key="different-key")

        request = self._create_guest_request(
            cart_id=existing_cart.id, session_key=self.session_key
        )

        cart_service = CartService(request=request)
        cart = cart_service.get_or_create_cart()

        self.assertNotEqual(cart.id, existing_cart.id)
        self.assertEqual(cart.session_key, self.session_key)

    def test_guest_cart_no_headers_creates_new(self):
        request = self._create_guest_request()

        cart_service = CartService(request=request)
        cart = cart_service.get_or_create_cart()

        self.assertIsNotNone(cart)
        self.assertIsNotNone(cart.session_key)
        self.assertIsNone(cart.user)

    def test_guest_cart_transition_to_user(self):
        guest_cart = CartFactory(user=None, session_key=self.session_key)
        CartItemFactory(cart=guest_cart, product=self.product, quantity=2)

        user = UserAccountFactory(num_addresses=0)

        guest_request = self._create_guest_request(
            cart_id=guest_cart.id, session_key=self.session_key
        )
        guest_service = CartService(request=guest_request)
        self.assertEqual(guest_service.cart.items.count(), 1)

        user_request = self._create_request_with_headers(
            user=user, cart_id=guest_cart.id, session_key=self.session_key
        )

        user_service = CartService(request=user_request)

        self.assertEqual(user_service.cart.user, user)
        self.assertEqual(user_service.cart.items.count(), 1)
        self.assertEqual(user_service.cart.items.first().quantity, 2)

        self.assertFalse(Cart.objects.filter(id=guest_cart.id).exists())

    def _create_request_with_headers(
        self, user, cart_id=None, session_key=None
    ):
        request = self.factory.get("/")
        request.user = user

        if cart_id:
            request.META["HTTP_X_CART_ID"] = str(cart_id)
        if session_key:
            request.META["HTTP_X_SESSION_KEY"] = session_key

        request.session = {}
        return request



from django.core.files.storage import storages
from storages.backends.s3boto3 import S3Boto3Storage


class StaticStorage(S3Boto3Storage):
    location = "static"
    default_acl = "public-read"

    def __init__(self, **settings):
        super().__init__(**settings)
        self.local_storage = storages.create_storage(
            {"BACKEND": "compressor.storage.CompressorFileStorage"}
        )

    def save(self, name, content, max_length=None):
        self.local_storage.save(name, content)
        super().save(name, self.local_storage._open(name))
        return name


class PublicMediaStorage(S3Boto3Storage):
    location = "media"
    default_acl = "public-read"
    file_overwrite = False


class PrivateMediaStorage(S3Boto3Storage):
    location = "private"
    default_acl = "private"
    file_overwrite = False
    custom_domain = False


class TinymceS3Storage(S3Boto3Storage):
    location = "media/uploads/tinymce"
    default_acl = "public-read"
    file_overwrite = False



# Generated by Django 5.0.3 on 2024-03-08 15:04
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0002_initial"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.AddField(
            model_name="settings",
            name="site",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="settings",
                to="sites.site",
                verbose_name="Site",
            ),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name="settings",
            unique_together={("site", "key")},
        ),
    ]



from django.test import TestCase

from core.utils.html import preprocess_text


class HtmlTest(TestCase):
    def test_preprocess_text(self):
        self.assertEqual(
            preprocess_text("<p>Hello&nbsp;World!</p>"), "Hello World"
        )
        self.assertEqual(
            preprocess_text("   Clean   spaces   "), "Clean spaces"
        )
        self.assertEqual(preprocess_text(""), "")
        self.assertEqual(preprocess_text("<a href='link'>Link</a>"), "Link")
        self.assertEqual(preprocess_text("&amp;"), "&")
        self.assertEqual(preprocess_text("Ελληνικά"), "Ελληνικά")



import uuid
from typing import Any, cast

from django.contrib.postgres.indexes import BTreeIndex, GinIndex
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, transaction
from django.db.models import F, JSONField, Max, Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta


class SortableModel(models.Model):
    sort_order = models.IntegerField(_("Sort Order"), editable=False, null=True)

    class Meta(TypedModelMeta):
        abstract = True
        indexes = [
            BTreeIndex(fields=["sort_order"], name="%(class)s_sort_order_ix"),
        ]

    def save(self, *args, **kwargs):
        if self.pk is None:
            qs = self.get_ordering_queryset()
            existing_max = self.get_max_sort_order(qs)
            self.sort_order = 0 if existing_max is None else existing_max + 1
        super().save(*args, **kwargs)

    def get_ordering_queryset(self):
        model_class = cast("type[models.Model]", self.__class__)
        return model_class._default_manager.all()

    @staticmethod
    def get_max_sort_order(qs):
        return qs.aggregate(Max("sort_order"))["sort_order__max"] or 0

    def move_up(self):
        if self.sort_order is not None and self.sort_order > 0:
            qs = self.get_ordering_queryset()
            prev_item = qs.get(sort_order=self.sort_order - 1)
            prev_item.sort_order, self.sort_order = (
                self.sort_order,
                prev_item.sort_order,
            )
            prev_item.save()
            self.save()

    def move_down(self):
        if self.sort_order is not None:
            qs = self.get_ordering_queryset()
            next_item = qs.filter(sort_order__gt=self.sort_order).first()
            if next_item:
                next_item.sort_order, self.sort_order = (
                    self.sort_order,
                    next_item.sort_order,
                )
                next_item.save()
                self.save()

    @transaction.atomic
    def delete(self, *args, **kwargs):
        if self.sort_order is not None:
            qs = self.get_ordering_queryset()
            qs.filter(sort_order__gt=self.sort_order).update(
                sort_order=F("sort_order") - 1
            )
        super().delete(*args, **kwargs)


class TimeStampMixinModel(models.Model):
    created_at = models.DateTimeField(_("Created At"), auto_now_add=True)
    updated_at = models.DateTimeField(_("Updated At"), auto_now=True)

    class Meta(TypedModelMeta):
        abstract = True
        indexes = [
            BTreeIndex(fields=["created_at"], name="%(class)s_created_at_ix"),
            BTreeIndex(fields=["updated_at"], name="%(class)s_updated_at_ix"),
        ]

    def get_duration_since_created(self):
        return timezone.now() - self.created_at

    def get_duration_since_updated(self):
        return timezone.now() - self.updated_at


class UUIDModel(models.Model):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    class Meta(TypedModelMeta):
        abstract = True


class PublishedQuerySet(models.QuerySet):
    def published(self):
        today = timezone.now()
        return self.filter(
            Q(published_at__lte=today, is_published=True)
            | Q(published_at__isnull=True, is_published=True)
        )


PublishableManager = models.Manager.from_queryset(PublishedQuerySet)


class PublishableModel(models.Model):
    published_at = models.DateTimeField(
        _("Published At"), null=True, blank=True
    )
    is_published = models.BooleanField(_("Is Published"), default=False)

    objects: PublishableManager = PublishableManager()

    class Meta(TypedModelMeta):
        abstract = True
        indexes = [
            BTreeIndex(
                fields=["published_at"], name="%(class)s_published_at_ix"
            ),
            BTreeIndex(
                fields=["is_published"], name="%(class)s_is_published_ix"
            ),
        ]

    def save(self, *args, **kwargs):
        if self.is_published and self.published_at is None:
            self.published_at = timezone.now()
        super().save(*args, **kwargs)


class MetaDataModel(models.Model):
    private_metadata = JSONField(
        blank=True, default=dict, encoder=DjangoJSONEncoder
    )
    metadata = JSONField(blank=True, default=dict, encoder=DjangoJSONEncoder)

    class Meta(TypedModelMeta):
        indexes = [
            GinIndex(fields=["private_metadata"], name="%(class)s_p_meta_ix"),
            GinIndex(fields=["metadata"], name="%(class)s_meta_ix"),
        ]
        abstract = True

    def save(self, *args, **kwargs):
        if not self.private_metadata:
            self.private_metadata = {}
        if not self.metadata:
            self.metadata = {}
        super().save(*args, **kwargs)

    def get_value_from_private_metadata(self, key: str, default: Any = None):
        return self.private_metadata.get(key, default)

    def store_value_in_private_metadata(self, items: dict):
        if items:
            for key, value in items.items():
                self.private_metadata[key] = value
            self.save(update_fields=["private_metadata"])
        self.refresh_from_db(fields=["private_metadata"])

    def clear_private_metadata(self):
        self.private_metadata = {}

    def delete_value_from_private_metadata(self, key: str):
        if key in self.private_metadata:
            del self.private_metadata[key]

    def get_value_from_metadata(self, key: str, default: Any = None):
        return self.metadata.get(key, default)

    def store_value_in_metadata(self, items: dict):
        if items:
            for key, value in items.items():
                self.metadata[key] = value
            self.save(update_fields=["metadata"])
        self.refresh_from_db(fields=["metadata"])

    def clear_metadata(self):
        self.metadata = {}

    def delete_value_from_metadata(self, key: str):
        if key in self.metadata:
            del self.metadata[key]


class SoftDeleteMixin(models.Model):
    deleted_at = models.DateTimeField(null=True, blank=True)
    is_deleted = models.BooleanField(default=False)

    class Meta(TypedModelMeta):
        abstract = True

    def delete(self, using=None, keep_parents=False):
        self.deleted_at = timezone.now()
        self.is_deleted = True
        self.save()

    def restore(self):
        self.deleted_at = None
        self.is_deleted = False
        self.save()


class SoftDeleteQuerySet(models.QuerySet):
    def delete(self):
        return super().update(deleted_at=timezone.now(), is_deleted=True)

    def restore(self):
        return super().update(deleted_at=None, is_deleted=False)

    def hard_delete(self):
        return super().delete()


class SoftDeleteManager(models.Manager):
    def get_queryset(self) -> SoftDeleteQuerySet:
        return SoftDeleteQuerySet(self.model, using=self._db).exclude(
            is_deleted=True
        )

    def all_with_deleted(self):
        return SoftDeleteQuerySet(self.model, using=self._db)

    def deleted_only(self):
        return SoftDeleteQuerySet(self.model, using=self._db).filter(
            is_deleted=True
        )


class SoftDeleteModel(SoftDeleteMixin, models.Model):
    class Meta(TypedModelMeta):
        abstract = True



from typing import Literal

from django.contrib.postgres.indexes import BTreeIndex
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta
from parler.models import TranslatableModel, TranslatedFields

from blog.managers.author import BlogAuthorManager
from core.models import TimeStampMixinModel, UUIDModel


class BlogAuthor(TranslatableModel, TimeStampMixinModel, UUIDModel):
    id = models.BigAutoField(primary_key=True)
    user = models.OneToOneField("user.UserAccount", on_delete=models.PROTECT)
    website = models.URLField(_("Website"), blank=True, default="")
    translations = TranslatedFields(
        bio=models.TextField(_("Bio"), blank=True, null=True)
    )

    objects: BlogAuthorManager = BlogAuthorManager()

    class Meta(TypedModelMeta):
        verbose_name = _("Blog Author")
        verbose_name_plural = _("Blog Authors")
        ordering = ["-created_at"]
        indexes = [
            *TimeStampMixinModel.Meta.indexes,
            BTreeIndex(fields=["user"], name="blog_author_user_ix"),
        ]

    def __str__(self):
        author_name = self.user.full_name
        return f"{author_name} ({self.user.email})"

    @property
    def full_name(self) -> str:
        return self.user.full_name

    @property
    def image(self) -> str:
        return self.user.image

    @property
    def number_of_posts(self) -> int:
        return self.blog_posts.count()

    @property
    def total_likes_received(self) -> int | Literal[0]:
        return sum([post.likes.count() for post in self.blog_posts.all()])



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from cart.factories.cart import CartFactory
from cart.factories.item import CartItemFactory
from cart.models import CartItem
from core.utils.testing import TestURLFixerMixin
from product.factories.product import ProductFactory
from user.factories.account import UserAccountFactory

User = get_user_model()


class CartItemViewSetTest(TestURLFixerMixin, APITestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = UserAccountFactory(num_addresses=0)
        cls.other_user = UserAccountFactory(num_addresses=0)

        cls.cart = CartFactory(user=cls.user, num_cart_items=0)
        cls.other_cart = CartFactory(user=cls.other_user, num_cart_items=0)

        cls.product1 = ProductFactory(
            active=True, num_images=0, num_reviews=0, stock=10
        )
        cls.product2 = ProductFactory(
            active=True, num_images=0, num_reviews=0, stock=5
        )

        cls.cart_item = CartItemFactory(
            cart=cls.cart, product=cls.product1, quantity=2
        )
        cls.other_cart_item = CartItemFactory(
            cart=cls.other_cart, product=cls.product2, quantity=3
        )

    def setUp(self):
        self.client.force_authenticate(user=self.user)
        self.list_url = reverse("cart-item-list")
        self.detail_url = reverse(
            "cart-item-detail", kwargs={"pk": self.cart_item.pk}
        )

    def test_list_uses_correct_serializer(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)
        self.assertEqual(len(response.data["results"]), 1)

        item_data = response.data["results"][0]

        expected_fields = {
            "id",
            "uuid",
            "cart_id",
            "product",
            "quantity",
            "price",
            "final_price",
            "discount_value",
            "total_price",
            "total_discount_value",
            "vat_value",
            "vat_percent",
            "discount_percent",
            "price_save_percent",
            "created_at",
            "updated_at",
        }
        self.assertTrue(expected_fields.issubset(set(item_data.keys())))

    def test_retrieve_uses_correct_serializer(self):
        response = self.client.get(self.detail_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "uuid",
            "cart_id",
            "product",
            "quantity",
            "price",
            "final_price",
            "discount_value",
            "total_price",
            "total_discount_value",
            "vat_value",
            "vat_percent",
            "discount_percent",
            "price_save_percent",
            "created_at",
            "updated_at",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_create_uses_correct_serializer(self):
        new_product = ProductFactory(active=True, num_images=0, num_reviews=0)
        create_data = {
            "product": new_product.pk,
            "quantity": 3,
        }

        response = self.client.post(self.list_url, create_data, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        expected_fields = {
            "id",
            "uuid",
            "cart_id",
            "product",
            "quantity",
            "price",
            "final_price",
            "discount_value",
            "total_price",
            "total_discount_value",
            "vat_value",
            "vat_percent",
            "discount_percent",
            "price_save_percent",
            "created_at",
            "updated_at",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_update_uses_correct_serializer(self):
        update_data = {"quantity": 5}

        response = self.client.patch(
            self.detail_url, update_data, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "uuid",
            "cart_id",
            "product",
            "quantity",
            "price",
            "final_price",
            "discount_value",
            "total_price",
            "total_discount_value",
            "vat_value",
            "vat_percent",
            "discount_percent",
            "price_save_percent",
            "created_at",
            "updated_at",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_list_cart_items(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)
        self.assertIn("count", response.data)

        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["id"], self.cart_item.id)

    def test_retrieve_cart_item(self):
        response = self.client.get(self.detail_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], self.cart_item.id)
        self.assertEqual(response.data["quantity"], self.cart_item.quantity)

    def test_create_cart_item(self):
        new_product = ProductFactory(active=True, num_images=0, num_reviews=0)
        create_data = {
            "product": new_product.pk,
            "quantity": 3,
        }

        response = self.client.post(self.list_url, create_data, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(
            CartItem.objects.filter(
                cart=self.cart, product=new_product, quantity=3
            ).exists()
        )

    def test_update_cart_item(self):
        update_data = {"quantity": 5}

        response = self.client.patch(
            self.detail_url, update_data, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.cart_item.refresh_from_db()
        self.assertEqual(self.cart_item.quantity, 5)

    def test_delete_cart_item(self):
        response = self.client.delete(self.detail_url)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertFalse(CartItem.objects.filter(id=self.cart_item.id).exists())

    def test_cannot_access_other_user_cart_items(self):
        other_item_url = reverse(
            "cart-item-detail", kwargs={"pk": self.other_cart_item.pk}
        )

        response = self.client.get(other_item_url)
        self.assertIn(
            response.status_code,
            [status.HTTP_403_FORBIDDEN, status.HTTP_404_NOT_FOUND],
        )

    def test_cannot_update_other_user_cart_items(self):
        other_item_url = reverse(
            "cart-item-detail", kwargs={"pk": self.other_cart_item.pk}
        )

        response = self.client.patch(
            other_item_url, {"quantity": 10}, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_cannot_delete_other_user_cart_items(self):
        other_item_url = reverse(
            "cart-item-detail", kwargs={"pk": self.other_cart_item.pk}
        )

        response = self.client.delete(other_item_url)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_cart_item_guest_user(self):
        self.client.force_authenticate(user=None)

        guest_cart = CartFactory(user=None, num_cart_items=0)
        product = ProductFactory(active=True, num_images=0, num_reviews=0)

        headers = {
            "HTTP_X_CART_ID": str(guest_cart.id),
            "HTTP_X_SESSION_KEY": guest_cart.session_key,
        }

        create_data = {
            "product": product.pk,
            "quantity": 2,
        }

        response = self.client.post(
            self.list_url, create_data, format="json", **headers
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(
            CartItem.objects.filter(
                cart=guest_cart, product=product, quantity=2
            ).exists()
        )

    def test_list_guest_cart_items(self):
        self.client.force_authenticate(user=None)

        guest_cart = CartFactory(user=None, num_cart_items=1)

        headers = {
            "HTTP_X_CART_ID": str(guest_cart.id),
            "HTTP_X_SESSION_KEY": guest_cart.session_key,
        }

        response = self.client.get(self.list_url, **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 1)

    def test_filter_by_product(self):
        response = self.client.get(self.list_url, {"product": self.product1.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(
            response.data["results"][0]["product"]["id"], self.product1.id
        )

    def test_filter_by_quantity_range(self):
        CartItemFactory(cart=self.cart, quantity=1)
        CartItemFactory(cart=self.cart, quantity=5)

        response = self.client.get(self.list_url, {"min_quantity": 2})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        quantities = [item["quantity"] for item in response.data["results"]]
        self.assertTrue(all(q >= 2 for q in quantities))

        response = self.client.get(self.list_url, {"max_quantity": 3})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        quantities = [item["quantity"] for item in response.data["results"]]
        self.assertTrue(all(q <= 3 for q in quantities))

    def test_filter_by_product_name(self):
        self.product1.set_current_language("en")
        self.product1.name = "Test Product Name"
        self.product1.save()

        response = self.client.get(
            self.list_url, {"product_name": "Test Product"}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 1)

    def test_ordering_by_quantity(self):
        CartItemFactory(cart=self.cart, quantity=1)
        CartItemFactory(cart=self.cart, quantity=5)

        response = self.client.get(self.list_url, {"ordering": "quantity"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        quantities = [item["quantity"] for item in response.data["results"]]
        self.assertEqual(quantities, sorted(quantities))

    def test_ordering_by_created_at_desc(self):
        response = self.client.get(self.list_url, {"ordering": "-created_at"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertGreater(len(response.data["results"]), 0)

    def test_search_by_product_name(self):
        self.product1.set_current_language("en")
        self.product1.name = "Searchable Product"
        self.product1.save()

        response = self.client.get(self.list_url, {"search": "Searchable"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 1)

    def test_create_with_invalid_quantity(self):
        create_data = {
            "product": self.product1.pk,
            "quantity": 0,
        }

        response = self.client.post(self.list_url, create_data, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_with_nonexistent_product(self):
        create_data = {
            "product": 99999,
            "quantity": 1,
        }

        response = self.client.post(self.list_url, create_data, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_update_with_invalid_quantity(self):
        update_data = {"quantity": -1}

        response = self.client.patch(
            self.detail_url, update_data, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retrieve_nonexistent_cart_item(self):
        nonexistent_url = reverse("cart-item-detail", kwargs={"pk": 99999})

        response = self.client.get(nonexistent_url)
        self.assertIn(
            response.status_code,
            [status.HTTP_403_FORBIDDEN, status.HTTP_404_NOT_FOUND],
        )

    def test_unauthenticated_access_without_guest_headers(self):
        self.client.force_authenticate(user=None)

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 0)

    def test_cart_item_price_calculations(self):
        response = self.client.get(self.detail_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        item_data = response.data
        self.assertIn("price", item_data)
        self.assertIn("final_price", item_data)
        self.assertIn("total_price", item_data)
        self.assertIn("discount_value", item_data)
        self.assertIn("vat_value", item_data)

    def test_duplicate_cart_item_handling(self):
        create_data = {
            "product": self.product1.pk,
            "quantity": 1,
        }

        response = self.client.post(self.list_url, create_data, format="json")
        self.assertIn(
            response.status_code,
            [
                status.HTTP_400_BAD_REQUEST,
                status.HTTP_201_CREATED,
            ],
        )



import importlib
import uuid

import factory
from django.apps import apps
from django.conf import settings
from faker import Faker

from blog.factories.comment import BlogCommentFactory
from blog.factories.tag import BlogTagFactory
from blog.models.post import BlogPost
from core.factories import CustomDjangoModelFactory


fake = Faker()

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]


def generate_unique_blog_filename():
    return f"blog_post_{uuid.uuid4().hex[:8]}.jpg"


def get_or_create_category():
    if apps.get_model("blog", "BlogCategory").objects.exists():
        return (
            apps.get_model("blog", "BlogCategory").objects.order_by("?").first()
        )
    else:
        category_factory_module = importlib.import_module(
            "blog.factories.category"
        )
        category_factory_class = category_factory_module.BlogCategoryFactory
        return category_factory_class.create()


def get_or_create_author():
    if apps.get_model("blog", "BlogAuthor").objects.exists():
        return (
            apps.get_model("blog", "BlogAuthor").objects.order_by("?").first()
        )
    else:
        author_factory_module = importlib.import_module("blog.factories.author")
        author_factory_class = author_factory_module.BlogAuthorFactory
        return author_factory_class.create()


class BlogPostTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    title = factory.Faker("sentence", nb_words=6)
    subtitle = factory.Faker("sentence", nb_words=12)
    body = factory.Faker("paragraph", nb_sentences=5)
    master = factory.SubFactory("blog.factories.post.BlogPostFactory")

    class Meta:
        model = apps.get_model("blog", "BlogPostTranslation")
        django_get_or_create = ("language_code", "master")


class BlogPostFactory(CustomDjangoModelFactory):
    auto_translations = False

    unique_model_fields = [
        ("slug", lambda: fake.slug()),
    ]

    image = factory.django.ImageField(
        filename=factory.LazyFunction(generate_unique_blog_filename),
        width=1200,
        height=630,
        color="purple",
    )
    category = factory.LazyFunction(get_or_create_category)
    author = factory.LazyFunction(get_or_create_author)
    is_published = factory.Faker("boolean")
    featured = factory.Faker("boolean")
    view_count = factory.Faker("random_int", min=0, max=1000)

    class Meta:
        model = BlogPost
        django_get_or_create = ("slug",)
        skip_postgeneration_save = True

    @factory.post_generation
    def num_tags(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            tags = BlogTagFactory.create_batch(extracted)
            self.tags.add(*tags)

    @factory.post_generation
    def num_comments(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            BlogCommentFactory.create_batch(extracted, post=self)

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        translations = extracted or [
            BlogPostTranslationFactory(language_code=lang, master=self)
            for lang in available_languages
        ]

        for translation in translations:
            translation.master = self
            translation.save()



from datetime import timedelta
from typing import Any

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from drf_spectacular.helpers import lazy_serializer
from drf_spectacular.utils import extend_schema_field
from mptt.fields import TreeForeignKey
from parler_rest.serializers import TranslatableModelSerializer
from rest_framework import serializers
from rest_framework.relations import PrimaryKeyRelatedField

from user.serializers.account import UserDetailsSerializer
from blog.models.comment import BlogComment
from blog.models.post import BlogPost
from core.api.schema import generate_schema_multi_lang
from core.utils.serializers import TranslatedFieldExtended

User = get_user_model()


@extend_schema_field(generate_schema_multi_lang(BlogComment))
class TranslatedFieldsFieldExtend(TranslatedFieldExtended):
    pass


class BlogCommentSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[BlogComment]
):
    user = UserDetailsSerializer(read_only=True)
    content_preview = serializers.SerializerMethodField(
        help_text=_("First 150 characters of the comment content")
    )
    is_reply = serializers.SerializerMethodField(
        help_text=_("Whether this comment is a reply to another comment")
    )
    has_replies = serializers.SerializerMethodField(
        help_text=_("Whether this comment has approved replies")
    )
    is_edited = serializers.SerializerMethodField(
        help_text=_("Whether this comment has been edited")
    )
    user_has_liked = serializers.SerializerMethodField(
        help_text=_("Whether the current user has liked this comment")
    )
    translations = TranslatedFieldsFieldExtend(shared_model=BlogComment)

    def get_content_preview(
        self, obj: BlogComment | TreeForeignKey
    ) -> str | None:
        content = obj.safe_translation_getter("content", any_language=True)
        if content:
            return content[:150] + "..." if len(content) > 150 else content
        return None

    def get_is_reply(self, obj: BlogComment) -> bool:
        return obj.parent is not None

    def get_has_replies(self, obj: BlogComment) -> bool:
        return obj.get_children().filter(approved=True).exists()

    def get_is_edited(self, obj: BlogComment) -> bool:
        return obj.updated_at > obj.created_at + timedelta(minutes=5)

    def get_user_has_liked(self, obj: BlogComment) -> bool:
        request = self.context.get("request")
        if request and request.user.is_authenticated:
            return obj.likes.filter(id=request.user.id).exists()
        return False

    class Meta:
        model = BlogComment
        fields = (
            "id",
            "translations",
            "user",
            "content_preview",
            "is_reply",
            "parent",
            "has_replies",
            "approved",
            "is_edited",
            "likes_count",
            "replies_count",
            "user_has_liked",
            "created_at",
            "updated_at",
            "uuid",
        )
        read_only_fields = (
            "id",
            "user",
            "content_preview",
            "is_reply",
            "parent",
            "has_replies",
            "approved",
            "is_edited",
            "likes_count",
            "replies_count",
            "user_has_liked",
            "created_at",
            "updated_at",
            "uuid",
        )


class BlogCommentDetailSerializer(BlogCommentSerializer):
    post = serializers.SerializerMethodField(
        help_text=_("Basic information about the blog post")
    )
    parent_comment = serializers.SerializerMethodField(
        help_text=_("Parent comment if this is a reply")
    )
    children_comments = serializers.SerializerMethodField(
        help_text=_("Direct approved child comments (replies)")
    )
    ancestors_path = serializers.SerializerMethodField(
        help_text=_("Path from root comment to this comment")
    )
    tree_position = serializers.SerializerMethodField(
        help_text=_("Position information in the comment tree")
    )

    @extend_schema_field(
        lazy_serializer("blog.serializers.post.BlogPostSerializer")()
    )
    def get_post(self, obj: BlogComment):
        from blog.serializers.post import BlogPostSerializer  # noqa: PLC0415, I001

        return BlogPostSerializer(obj.post, context=self.context).data

    @extend_schema_field(
        {
            "type": "object",
            "nullable": True,
            "properties": {
                "id": {"type": "integer"},
                "content_preview": {"type": "string"},
                "user": {"$ref": "#/components/schemas/UserDetails"},
                "created_at": {"type": "string", "format": "date-time"},
            },
            "required": ["id", "content_preview", "user", "created_at"],
        }
    )
    def get_parent_comment(self, obj: BlogComment):
        if obj.parent:
            return {
                "id": obj.parent.id,
                "content_preview": self.get_content_preview(obj.parent),
                "user": UserDetailsSerializer(obj.parent.user).data,
                "created_at": obj.parent.created_at,
            }
        return None

    @extend_schema_field(
        lazy_serializer("blog.serializers.comment.BlogCommentSerializer")(
            many=True
        )
    )
    def get_children_comments(self, obj: BlogComment):
        request = self.context.get("request")
        children = (
            obj.get_children().select_related("user").order_by("created_at")
        )

        if not (request and request.user.is_staff):
            children = children.filter(approved=True)

        if children.exists():
            return BlogCommentSerializer(
                children, many=True, context=self.context
            ).data
        return []

    @extend_schema_field(
        {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {"type": "integer"},
                    "content_preview": {"type": "string"},
                    "user": {"$ref": "#/components/schemas/UserDetails"},
                },
                "required": ["id", "content_preview", "user"],
            },
        }
    )
    def get_ancestors_path(self, obj: BlogComment):
        request = self.context.get("request")
        ancestors = obj.get_ancestors().select_related("user")

        if not (request and request.user.is_staff):
            ancestors = ancestors.filter(approved=True)

        return [
            {
                "id": ancestor.id,
                "content_preview": self.get_content_preview(ancestor),
                "user": UserDetailsSerializer(ancestor.user).data,
            }
            for ancestor in ancestors
        ]

    @extend_schema_field(
        {
            "type": "object",
            "properties": {
                "level": {"type": "integer"},
                "tree_id": {"type": "integer"},
                "approved_descendants_count": {"type": "integer"},
                "siblings_count": {"type": "integer"},
            },
            "required": [
                "level",
                "tree_id",
                "approved_descendants_count",
                "siblings_count",
            ],
        }
    )
    def get_tree_position(self, obj: BlogComment):
        return {
            "level": obj.level,
            "tree_id": obj.tree_id,
            "approved_descendants_count": obj.approved_descendants_count,
            "siblings_count": obj.get_siblings().filter(approved=True).count(),
        }

    class Meta(BlogCommentSerializer.Meta):
        fields = (
            *BlogCommentSerializer.Meta.fields,
            "post",
            "parent_comment",
            "children_comments",
            "ancestors_path",
            "tree_position",
        )


class BlogCommentWriteSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[BlogComment]
):
    user = PrimaryKeyRelatedField(queryset=User.objects.all(), required=False)
    post = PrimaryKeyRelatedField(queryset=BlogPost.objects.all())
    parent = PrimaryKeyRelatedField(
        queryset=BlogComment.objects.all(), required=False, allow_null=True
    )
    translations = TranslatedFieldsFieldExtend(shared_model=BlogComment)

    def validate_parent(self, value: BlogComment) -> BlogComment:
        if value:
            post = self.initial_data.get("post")
            if isinstance(post, int):
                if value.post.id != post:
                    raise serializers.ValidationError(
                        _("Parent comment must belong to the same post.")
                    )
            elif post and value.post != post:
                raise serializers.ValidationError(
                    _("Parent comment must belong to the same post.")
                )
        return value

    def create(self, validated_data: Any) -> BlogComment:
        if "user" not in validated_data:
            validated_data["user"] = self.context["request"].user
        return super().create(validated_data)

    class Meta:
        model = BlogComment
        fields = (
            "translations",
            "user",
            "post",
            "parent",
        )
        read_only_fields = ("user",)


class BlogCommentLikedCommentsRequestSerializer(serializers.Serializer):
    comment_ids = serializers.ListField(
        child=serializers.IntegerField(),
        help_text=_("List of comment IDs to check like status for"),
    )

    def validate_comment_ids(self, value: list) -> list:
        if not value:
            raise serializers.ValidationError(
                _("At least one comment ID is required.")
            )
        if len(value) > 100:
            raise serializers.ValidationError(
                _("Cannot check more than 100 comments at once.")
            )
        return value


class BlogCommentLikedCommentsResponseSerializer(serializers.Serializer):
    liked_comment_ids = serializers.ListField(
        child=serializers.IntegerField(),
        help_text=_("List of comment IDs that are liked by the current user"),
    )



from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from tag.factories.tag import TagFactory
from tag.factories.tagged_item import TaggedProductFactory
from tag.models.tagged_item import TaggedItem
from product.factories.product import ProductFactory
from core.utils.testing import TestURLFixerMixin


class TaggedItemViewSetTestCase(TestURLFixerMixin, APITestCase):
    @classmethod
    def setUpTestData(cls):
        cls.tag = TagFactory(active=True)
        cls.inactive_tag = TagFactory(active=False)
        cls.product = ProductFactory()
        cls.tagged_item = TaggedProductFactory(
            tag=cls.tag, content_object=cls.product
        )

    def get_tagged_item_detail_url(self, pk):
        return reverse("tagged-item-detail", args=[pk])

    def get_tagged_item_list_url(self):
        return reverse("tagged-item-list")

    def test_list(self):
        url = self.get_tagged_item_list_url()
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)
        self.assertIn("count", response.data)
        if "links" in response.data:
            self.assertIn("next", response.data["links"])
            self.assertIn("previous", response.data["links"])
        else:
            self.assertIn("next", response.data)
            self.assertIn("previous", response.data)

        self.assertGreaterEqual(len(response.data["results"]), 1)

        first_result = response.data["results"][0]
        expected_basic_fields = {
            "id",
            "tag",
            "content_type",
            "content_type_name",
            "object_id",
            "content_object",
            "created_at",
            "updated_at",
            "uuid",
        }
        self.assertTrue(
            expected_basic_fields.issubset(set(first_result.keys()))
        )

    def test_retrieve_valid(self):
        url = self.get_tagged_item_detail_url(self.tagged_item.id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "tag",
            "content_type",
            "content_type_name",
            "object_id",
            "content_object",
            "created_at",
            "updated_at",
            "uuid",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_retrieve_invalid(self):
        invalid_tagged_item_id = 9999
        url = self.get_tagged_item_detail_url(invalid_tagged_item_id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_valid(self):
        new_product = ProductFactory()
        content_type = ContentType.objects.get_for_model(new_product)

        payload = {
            "tag_id": self.tag.id,
            "content_type": content_type.id,
            "object_id": new_product.id,
        }

        url = self.get_tagged_item_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        expected_fields = {
            "id",
            "tag",
            "content_type",
            "content_type_name",
            "object_id",
            "content_object",
            "created_at",
            "updated_at",
            "uuid",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_create_invalid_tag(self):
        new_product = ProductFactory()
        content_type = ContentType.objects.get_for_model(new_product)

        payload = {
            "tag_id": 99999,
            "content_type": content_type.id,
            "object_id": new_product.id,
        }

        url = self.get_tagged_item_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("tag_id", response.data)

    def test_create_inactive_tag(self):
        new_product = ProductFactory()
        content_type = ContentType.objects.get_for_model(new_product)

        payload = {
            "tag_id": self.inactive_tag.id,
            "content_type": content_type.id,
            "object_id": new_product.id,
        }

        url = self.get_tagged_item_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("tag_id", response.data)

    def test_update_valid(self):
        new_tag = TagFactory(active=True)
        payload = {
            "tag_id": new_tag.id,
            "content_type": self.tagged_item.content_type.id,
            "object_id": self.tagged_item.object_id,
        }

        url = self.get_tagged_item_detail_url(self.tagged_item.id)
        response = self.client.put(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "tag",
            "content_type",
            "content_type_name",
            "object_id",
            "content_object",
            "created_at",
            "updated_at",
            "uuid",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_destroy_valid(self):
        tagged_item_to_delete = TaggedProductFactory(
            tag=self.tag, content_object=ProductFactory()
        )

        url = self.get_tagged_item_detail_url(tagged_item_to_delete.id)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertFalse(
            TaggedItem.objects.filter(id=tagged_item_to_delete.id).exists()
        )

    def test_destroy_invalid(self):
        invalid_tagged_item_id = 9999

        url = self.get_tagged_item_detail_url(invalid_tagged_item_id)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_filtering_by_tag(self):
        url = self.get_tagged_item_list_url()
        response = self.client.get(url, {"tag": self.tag.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for tagged_item_data in response.data["results"]:
            self.assertEqual(tagged_item_data["tag"]["id"], self.tag.id)

    def test_filtering_by_content_type(self):
        url = self.get_tagged_item_list_url()
        response = self.client.get(url, {"content_type": "product"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for tagged_item_data in response.data["results"]:
            self.assertEqual(tagged_item_data["content_type_name"], "product")

    def test_filtering_by_object_id(self):
        url = self.get_tagged_item_list_url()
        response = self.client.get(url, {"object_id": self.product.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for tagged_item_data in response.data["results"]:
            self.assertEqual(tagged_item_data["object_id"], self.product.id)

    def test_filtering_by_tag_label(self):
        url = self.get_tagged_item_list_url()
        response = self.client.get(url, {"tag__label": "test"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)

    def test_ordering_functionality(self):
        url = self.get_tagged_item_list_url()
        response = self.client.get(url, {"ordering": "created_at"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)

    def test_search_functionality(self):
        url = self.get_tagged_item_list_url()
        response = self.client.get(url, {"search": "test"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)



from __future__ import annotations

import os
import uuid
from decimal import Decimal

from django.conf import settings
from django.contrib.postgres.indexes import BTreeIndex
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import (
    Avg,
    F,
)
from django.utils.safestring import SafeString, mark_safe
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta
from djmoney.models.fields import MoneyField
from djmoney.money import Money
from measurement.measures import Weight
from mptt.fields import TreeForeignKey
from parler.fields import TranslationsForeignKey
from parler.models import TranslatableModel, TranslatedFieldsModel
from simple_history.models import HistoricalRecords
from tinymce.models import HTMLField

from core.fields.measurement import MeasurementField
from core.models import (
    MetaDataModel,
    SoftDeleteModel,
    TimeStampMixinModel,
    UUIDModel,
)
from core.units import WeightUnits
from core.utils.generators import SlugifyConfig, unique_slugify
from core.weight import zero_weight
from meili.models import IndexMixin
from product.managers.product import ProductManager
from product.models.favourite import ProductFavourite
from product.models.image import ProductImage
from product.models.review import ProductReview
from seo.models import SeoModel
from tag.models.tagged_item import TaggedModel

DISCOUNT_PERCENT_MIN = Decimal("0.0")
DISCOUNT_PERCENT_MAX = Decimal("100.0")


class Product(
    SoftDeleteModel,
    TranslatableModel,
    TimeStampMixinModel,
    SeoModel,
    UUIDModel,
    MetaDataModel,
    TaggedModel,
):
    id = models.BigAutoField(primary_key=True)
    sku = models.CharField(
        _("SKU"), unique=True, max_length=100, default=uuid.uuid4
    )
    category = TreeForeignKey(
        "product.ProductCategory",
        on_delete=models.SET_NULL,
        related_name="products",
        null=True,
        blank=True,
    )
    slug = models.SlugField(_("Slug"), max_length=255, unique=True)
    price = MoneyField(
        _("Price"),
        max_digits=11,
        decimal_places=2,
        default=0,
    )
    active = models.BooleanField(_("Active"), default=True)
    stock = models.PositiveIntegerField(_("Stock"), default=0)
    discount_percent = models.DecimalField(
        _("Discount Percent"),
        max_digits=11,
        decimal_places=2,
        default=Decimal("0.0"),
    )
    vat = models.ForeignKey(
        "vat.Vat",
        related_name="products",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    view_count = models.PositiveBigIntegerField(_("View Count"), default=0)
    weight = MeasurementField(
        str(_("Weight")),
        measurement=Weight,
        unit_choices=WeightUnits.CHOICES,
        default=zero_weight,
    )
    changed_by = models.ForeignKey(
        "user.UserAccount",
        on_delete=models.SET_NULL,
        related_name="changed_products",
        null=True,
        blank=True,
    )
    history = HistoricalRecords()

    objects: ProductManager = ProductManager()

    class Meta(MetaDataModel.Meta, TypedModelMeta):
        verbose_name = _("Product")
        verbose_name_plural = _("Products")
        ordering = ["-created_at"]
        indexes = [
            *MetaDataModel.Meta.indexes,
            *TimeStampMixinModel.Meta.indexes,
            BTreeIndex(
                fields=["price", "stock"], name="product_price_stock_ix"
            ),
            BTreeIndex(fields=["sku"], name="product_sku_ix"),
            BTreeIndex(fields=["slug"], name="product_slug_ix"),
            BTreeIndex(fields=["price"], name="product_price_ix"),
            BTreeIndex(fields=["stock"], name="product_stock_ix"),
            BTreeIndex(fields=["discount_percent"], name="product_discount_ix"),
            BTreeIndex(fields=["view_count"], name="product_view_count_ix"),
            BTreeIndex(fields=["weight"], name="product_weight_ix"),
            BTreeIndex(fields=["active"], name="product_active_ix"),
            BTreeIndex(fields=["category"], name="product_category_ix"),
            BTreeIndex(
                fields=["active", "price"], name="product_active_price_ix"
            ),
        ]

    def __str__(self):
        return self.safe_translation_getter("name", any_language=True) or ""

    def __repr__(self):
        name = self.safe_translation_getter("name", any_language=True) or ""
        return f"<Product: {name} ({self.sku})>"

    def save(self, *args, **kwargs):
        if not self.sku:
            self.sku = self.generate_unique_sku()

        if not self.slug:
            config = SlugifyConfig(
                instance=self,
                title_field="name",
            )
            self.slug = unique_slugify(config)
        super().save(*args, **kwargs)

    def clean(self):
        super().clean()
        if self.discount_percent > 0 >= self.price.amount:
            raise ValidationError(
                {
                    "discount_percent": _(
                        "Discount percent cannot be greater than 0 if price is 0."
                    )
                }
            )

        if (
            not DISCOUNT_PERCENT_MIN
            <= self.discount_percent
            <= DISCOUNT_PERCENT_MAX
        ):
            raise ValidationError(
                {
                    "discount_percent": _(
                        "Discount percent must be between 0 and 100."
                    )
                }
            )

        if self.stock < 0:
            raise ValidationError({"stock": _("Stock cannot be negative.")})

    def generate_unique_sku(self) -> str:
        while True:
            unique_code = uuid.uuid4()
            if not self.objects.filter(sku=unique_code).exists():
                return str(unique_code)

    def increment_stock(self, quantity: int) -> None:
        if quantity < 0:
            raise ValueError("Quantity to increment must be non-negative")
        Product.objects.filter(id=self.id).update(stock=F("stock") + quantity)
        self.refresh_from_db()

    def decrement_stock(self, quantity: int) -> None:
        if quantity < 0:
            raise ValueError("Invalid quantity to decrement")
        updated_rows = Product.objects.filter(
            id=self.id, stock__gte=quantity
        ).update(stock=F("stock") - quantity)
        if not updated_rows:
            raise ValueError("Not enough stock to decrement")
        self.refresh_from_db()

    @property
    def _history_user(self):
        return self.changed_by

    @_history_user.setter
    def _history_user(self, value):
        self.changed_by = value

    @property
    def discount_value(self) -> Money:
        value = (self.price.amount * self.discount_percent) / 100
        return Money(value, settings.DEFAULT_CURRENCY)

    @property
    def price_save_percent(self) -> Decimal:
        if self.price.amount > 0:
            return Decimal(
                (self.discount_value.amount / self.price.amount) * 100
            )
        return Decimal(0)

    @property
    def likes_count(self) -> int:
        return ProductFavourite.objects.filter(product=self).count()

    @property
    def review_average(self) -> float:
        average = ProductReview.objects.filter(product=self).aggregate(
            avg=Avg("rate")
        )["avg"]
        return float(average) if average is not None else 0.0

    @property
    def review_count(self) -> int:
        return ProductReview.objects.filter(product=self).count()

    @property
    def vat_percent(self) -> Decimal:
        if self.vat:
            return self.vat.value
        return Decimal(0)

    @property
    def vat_value(self) -> Money:
        if self.vat:
            value = (self.price.amount * self.vat.value) / 100
            return Money(value, settings.DEFAULT_CURRENCY)
        return Money(0, settings.DEFAULT_CURRENCY)

    @property
    def final_price(self) -> Money:
        price_currency = self.price.currency
        vat_value = (
            Money(self.vat_value.amount, price_currency)
            if self.vat_value.currency != price_currency
            else self.vat_value
        )
        discount_value = (
            Money(self.discount_value.amount, price_currency)
            if self.discount_value.currency != price_currency
            else self.discount_value
        )
        return self.price + vat_value - discount_value

    @property
    def main_image_path(self) -> str:
        product_image = ProductImage.objects.filter(
            product_id=self.id, is_main=True
        ).first()
        if not product_image:
            return ""
        return f"media/uploads/products/{os.path.basename(product_image.image.name)}"

    @property
    def colored_stock(self) -> SafeString:
        if self.stock > 0:
            return mark_safe(
                '<span style="color: #1bff00;">{}</span>'.format(self.stock)
            )
        else:
            return mark_safe(
                '<span style="color: #ff0000;">{}</span>'.format(self.stock)
            )


class ProductTranslation(TranslatedFieldsModel, IndexMixin):
    master = TranslationsForeignKey(
        "product.Product",
        on_delete=models.CASCADE,
        related_name="translations",
        null=True,
    )
    name = models.CharField(_("Name"), max_length=255, blank=True, default="")
    description = HTMLField(_("Description"), blank=True, null=True)

    class Meta:
        app_label = "product"
        db_table = "product_product_translation"
        unique_together = ("language_code", "master")
        verbose_name = _("Product Translation")
        verbose_name_plural = _("Product Translations")

    class MeiliMeta:
        filterable_fields = (
            "name",
            "language_code",
            "likes_count",
            "final_price",
            "view_count",
            "category",
        )
        searchable_fields = ("id", "name", "description")
        displayed_fields = (
            "id",
            "name",
            "description",
            "language_code",
            "likes_count",
            "final_price",
            "view_count",
            "category",
        )
        sortable_fields = (
            "likes_count",
            "final_price",
            "view_count",
            "discount_percent",
        )
        ranking_rules = [
            "words",
            "typo",
            "proximity",
            "attribute",
            "sort",
            "likes_count:desc",
            "view_count:desc",
            "final_price:desc",
            "exactness",
        ]
        synonyms: dict = {}
        typo_tolerance = {
            "enabled": True,
            "minWordSizeForTypos": {"oneTypo": 3, "twoTypos": 5},
            "disableOnWords": [],
            "disableOnAttributes": [],
        }
        faceting = {"maxValuesPerFacet": 100}
        pagination = {"maxTotalHits": 1000}

    @classmethod
    def get_additional_meili_fields(cls):
        return {
            "likes_count": lambda obj: obj.master.likes_count,
            "view_count": lambda obj: obj.master.view_count,
            "final_price": lambda obj: float(obj.master.final_price.amount),
        }

    def __str__(self):
        name = self.name or "Untitled"
        return f"{name} ({self.language_code})"



from datetime import timedelta

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.utils import timezone

from user.models.subscription import SubscriptionTopic
from user.utils.subscription import send_newsletter

User = get_user_model()


class Command(BaseCommand):
    help = "Send weekly newsletter to all subscribers"

    def handle(self, *args, **options):
        self.stdout.write("Starting weekly newsletter send...")
        self.send_using_utility()

    def send_using_utility(self):
        try:
            topic = SubscriptionTopic.objects.get(slug="weekly-newsletter")
        except SubscriptionTopic.DoesNotExist:
            self.stdout.write(
                self.style.ERROR("Weekly newsletter topic not found!")
            )
            return

        context = {
            "week_start": timezone.now().date() - timedelta(days=7),
            "week_end": timezone.now().date(),
            "featured_articles": self.get_featured_articles(),
            "upcoming_events": self.get_upcoming_events(),
            "user_stats": self.get_user_stats(),
        }

        stats = send_newsletter(
            topic=topic,
            subject=f"Weekly Digest - Week {timezone.now().isocalendar()[1]}",
            template_name="emails/newsletter.html",
            context=context,
            batch_size=100,
        )

        self.stdout.write(
            self.style.SUCCESS(
                f"Newsletter sent successfully!\n"
                f"  - Sent: {stats['sent']}\n"
                f"  - Failed: {stats['failed']}\n"
                f"  - Skipped: {stats['skipped']}"
            )
        )

    def get_featured_articles(self):
        # @TODO - Add logic
        return []

    def get_upcoming_events(self):
        # @TODO - Add logic
        return []

    def get_user_stats(self):
        return {
            "new_users_this_week": User.objects.filter(
                created_at__gte=timezone.now() - timedelta(days=7)
            ).count(),
            "total_users": User.objects.count(),
        }



from django.db.models import Q, Count
from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as filters

from blog.models import BlogPost
from core.filters.camel_case_filters import (
    CamelCasePublishableTimeStampFilterSet,
)
from core.filters.core import UUIDFilterMixin


class BlogPostFilter(
    UUIDFilterMixin,
    CamelCasePublishableTimeStampFilterSet,
):
    title = filters.CharFilter(
        field_name="translations__title",
        lookup_expr="icontains",
        help_text=_("Filter by title (case-insensitive)"),
    )
    author_email = filters.CharFilter(
        field_name="author__user__email",
        lookup_expr="icontains",
        help_text=_("Filter by author email (case-insensitive)"),
    )
    min_likes = filters.NumberFilter(
        method="filter_min_likes",
        label="Minimum Likes Count",
        help_text=_("Filter by minimum number of likes"),
    )
    min_comments = filters.NumberFilter(
        method="filter_min_comments",
        label="Minimum Comments Count",
        help_text=_("Filter by minimum number of approved comments"),
    )
    min_tags = filters.NumberFilter(
        method="filter_min_tags",
        label="Minimum Tags Count",
        help_text=_("Filter by minimum number of active tags"),
    )
    featured = filters.BooleanFilter(
        field_name="featured",
        lookup_expr="exact",
        help_text=_("Filter by featured status"),
    )
    min_view_count = filters.NumberFilter(
        field_name="view_count",
        lookup_expr="gte",
        label="Minimum View Count",
        help_text=_("Filter by minimum number of views"),
    )
    category_name = filters.CharFilter(
        field_name="category__translations__name",
        lookup_expr="icontains",
        help_text=_("Filter by category name (case-insensitive)"),
    )
    tag_name = filters.CharFilter(
        field_name="tags__translations__label",
        lookup_expr="icontains",
        help_text=_("Filter by tag label (case-insensitive)"),
    )
    author_name = filters.CharFilter(
        method="filter_author_name",
        help_text=_("Filter by author full name (case-insensitive)"),
    )

    category = filters.ModelChoiceFilter(
        field_name="category",
        queryset=None,
        help_text=_("Filter by category ID"),
    )
    author = filters.ModelChoiceFilter(
        field_name="author",
        queryset=None,
        help_text=_("Filter by author ID"),
    )
    tags = filters.ModelMultipleChoiceFilter(
        field_name="tags",
        queryset=None,
        help_text=_("Filter by tag IDs (comma-separated)"),
    )

    class Meta:
        model = BlogPost
        fields = {
            "created_at": ["gte", "lte", "date"],
            "updated_at": ["gte", "lte", "date"],
            "published_at": ["gte", "lte", "date"],
            "is_published": ["exact"],
            "uuid": ["exact"],
            "id": ["exact", "in"],
            "slug": ["exact", "icontains"],
            "featured": ["exact"],
            "view_count": ["gte", "lte", "exact"],
            "category": ["exact"],
            "author": ["exact"],
            "tags": ["exact"],
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from blog.models import BlogCategory, BlogAuthor, BlogTag

        self.filters["category"].queryset = BlogCategory.objects.all()
        self.filters["author"].queryset = BlogAuthor.objects.all()
        self.filters["tags"].queryset = BlogTag.objects.all()

    def filter_author_name(self, queryset, name, value):
        """Filter posts by author full name (case-insensitive)."""
        if value:
            return queryset.filter(
                Q(author__user__first_name__icontains=value)
                | Q(author__user__last_name__icontains=value)
                | Q(author__user__username__icontains=value)
            )
        return queryset

    def filter_min_likes(self, queryset, name, value):
        """Filter posts with minimum number of likes."""
        if value is not None:
            return queryset.annotate(
                likes_count_annotation=Count("likes", distinct=True)
            ).filter(likes_count_annotation__gte=value)
        return queryset

    def filter_min_comments(self, queryset, name, value):
        """Filter posts with minimum number of approved comments."""
        if value is not None:
            return queryset.annotate(
                comments_count_annotation=Count(
                    "comments", distinct=True, filter=Q(comments__approved=True)
                )
            ).filter(comments_count_annotation__gte=value)
        return queryset

    def filter_min_tags(self, queryset, name, value):
        """Filter posts with minimum number of active tags."""
        if value is not None:
            return queryset.annotate(
                tags_count_annotation=Count(
                    "tags", distinct=True, filter=Q(tags__active=True)
                )
            ).filter(tags_count_annotation__gte=value)
        return queryset



# Generated by Django 5.1.4 on 2024-12-31 00:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('user', '0011_alter_useraccount_country_alter_useraccount_region_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='useraccount',
            name='address',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Address'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='bio',
            field=models.TextField(blank=True, default='', verbose_name='Bio'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='city',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='City'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='facebook',
            field=models.URLField(blank=True, default='', verbose_name='Facebook Profile'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='first_name',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='First Name'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='github',
            field=models.URLField(blank=True, default='', verbose_name='Github Profile'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='instagram',
            field=models.URLField(blank=True, default='', verbose_name='Instagram Profile'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='last_name',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Last Name'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='linkedin',
            field=models.URLField(blank=True, default='', verbose_name='LinkedIn Profile'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='place',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Place'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='twitter',
            field=models.URLField(blank=True, default='', verbose_name='Twitter Profile'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='website',
            field=models.URLField(blank=True, default='', verbose_name='Website'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='youtube',
            field=models.URLField(blank=True, default='', verbose_name='Youtube Profile'),
        ),
        migrations.AlterField(
            model_name='useraccount',
            name='zipcode',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Zip Code'),
        ),
        migrations.AlterField(
            model_name='useraddress',
            name='notes',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Notes'),
        ),
    ]



from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema_field
from parler_rest.serializers import TranslatableModelSerializer
from rest_framework import serializers

from blog.models.tag import BlogTag
from core.api.schema import generate_schema_multi_lang
from core.utils.serializers import TranslatedFieldExtended


@extend_schema_field(generate_schema_multi_lang(BlogTag))
class TranslatedFieldsFieldExtend(TranslatedFieldExtended):
    pass


class BlogTagSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[BlogTag]
):
    posts_count = serializers.CharField(
        source="get_posts_count",
        read_only=True,
        help_text=_("Number of blog posts using this tag"),
    )
    translations = TranslatedFieldsFieldExtend(shared_model=BlogTag)

    class Meta:
        model = BlogTag
        fields = (
            "id",
            "translations",
            "active",
            "sort_order",
            "posts_count",
            "created_at",
            "updated_at",
            "uuid",
        )
        read_only_fields = (
            "id",
            "posts_count",
            "created_at",
            "updated_at",
            "uuid",
        )


class BlogTagDetailSerializer(BlogTagSerializer):
    class Meta(BlogTagSerializer.Meta):
        fields = (*BlogTagSerializer.Meta.fields,)


class BlogTagWriteSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[BlogTag]
):
    translations = TranslatedFieldsFieldExtend(shared_model=BlogTag)

    def validate_sort_order(self, value: int) -> int:
        if value is not None and value < 0:
            raise serializers.ValidationError(
                _("Sort order cannot be negative.")
            )
        return value

    class Meta:
        model = BlogTag
        fields = (
            "translations",
            "active",
            "sort_order",
        )



from datetime import timedelta

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase

from core.utils.testing import TestURLFixerMixin
from notification.enum import (
    NotificationKindEnum,
)
from notification.factories.notification import NotificationFactory
from notification.factories.user import NotificationUserFactory
from notification.models.user import NotificationUser
from notification.serializers.user import (
    NotificationUserDetailSerializer,
    NotificationUserSerializer,
)
from user.factories.account import UserAccountFactory

User = get_user_model()


class NotificationUserViewSetTestCase(TestURLFixerMixin, APITestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = UserAccountFactory()
        cls.other_user = UserAccountFactory()
        cls.notification = NotificationFactory()
        cls.other_notification = NotificationFactory()

    def setUp(self):
        self.list_url = reverse("notification-user-list")

    def test_list_uses_correct_serializer(self):
        response = self.client.get(self.list_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("results", response.data)

        if response.data["results"]:
            serializer = NotificationUserSerializer(
                instance=response.data["results"][0]
            )
            expected_fields = set(serializer.data.keys())
            actual_fields = set(response.data["results"][0].keys())
            self.assertEqual(expected_fields, actual_fields)

    def test_retrieve_uses_correct_serializer(self):
        notification_user = NotificationUserFactory(
            user=self.user, notification=self.notification
        )
        detail_url = reverse(
            "notification-user-detail", args=[notification_user.id]
        )
        response = self.client.get(detail_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        serializer = NotificationUserDetailSerializer(
            instance=notification_user
        )
        expected_fields = set(serializer.data.keys())
        actual_fields = set(response.data.keys())
        self.assertEqual(expected_fields, actual_fields)

    def test_create_uses_correct_serializer(self):
        data = {
            "user": self.user.id,
            "notification": self.notification.id,
            "seen": False,
        }
        response = self.client.post(self.list_url, data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_obj = NotificationUser.objects.get(id=response.data["id"])
        serializer = NotificationUserDetailSerializer(instance=created_obj)
        expected_fields = set(serializer.data.keys())
        actual_fields = set(response.data.keys())
        self.assertEqual(expected_fields, actual_fields)

    def test_update_uses_correct_serializer(self):
        notification_user = NotificationUserFactory(
            user=self.user, notification=self.notification
        )
        detail_url = reverse(
            "notification-user-detail", args=[notification_user.id]
        )
        data = {
            "user": self.user.id,
            "notification": self.notification.id,
            "seen": True,
        }
        response = self.client.put(detail_url, data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["seen"], True)

    def test_partial_update_uses_correct_serializer(self):
        notification_user = NotificationUserFactory(
            user=self.user, notification=self.notification, seen=False
        )
        detail_url = reverse(
            "notification-user-detail", args=[notification_user.id]
        )
        data = {"seen": True}
        response = self.client.patch(detail_url, data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["seen"], True)

    def test_create_notification_user(self):
        initial_count = NotificationUser.objects.count()
        data = {
            "user": self.user.id,
            "notification": self.notification.id,
            "seen": False,
        }
        response = self.client.post(self.list_url, data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(NotificationUser.objects.count(), initial_count + 1)
        self.assertTrue(
            NotificationUser.objects.filter(
                user=self.user, notification=self.notification
            ).exists()
        )

    def test_create_duplicate_notification_user_fails(self):
        NotificationUserFactory(user=self.user, notification=self.notification)

        data = {
            "user": self.user.id,
            "notification": self.notification.id,
            "seen": False,
        }
        response = self.client.post(self.list_url, data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_list_notification_users(self):
        NotificationUserFactory(user=self.user, notification=self.notification)

        response = self.client.get(self.list_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("results", response.data)
        self.assertGreaterEqual(len(response.data["results"]), 1)

    def test_retrieve_notification_user(self):
        notification_user = NotificationUserFactory(
            user=self.user, notification=self.notification
        )
        detail_url = reverse(
            "notification-user-detail", args=[notification_user.id]
        )

        response = self.client.get(detail_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], notification_user.id)

    def test_update_notification_user(self):
        notification_user = NotificationUserFactory(
            user=self.user, notification=self.notification, seen=False
        )
        detail_url = reverse(
            "notification-user-detail", args=[notification_user.id]
        )

        data = {
            "user": self.user.id,
            "notification": self.notification.id,
            "seen": True,
        }
        response = self.client.put(detail_url, data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["seen"], True)

    def test_delete_notification_user(self):
        notification_user = NotificationUserFactory(
            user=self.user, notification=self.notification
        )
        detail_url = reverse(
            "notification-user-detail", args=[notification_user.id]
        )

        response = self.client.delete(detail_url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(
            NotificationUser.objects.filter(id=notification_user.id).exists()
        )

    def test_retrieve_nonexistent_notification_user(self):
        detail_url = reverse("notification-user-detail", args=[999999])
        response = self.client.get(detail_url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_filter_by_seen(self):
        seen_notification = NotificationUserFactory(
            user=self.user, notification=self.notification, seen=True
        )
        unseen_notification = NotificationUserFactory(
            user=self.user, notification=self.other_notification, seen=False
        )

        response = self.client.get(self.list_url, {"seen": "true"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        seen_ids = [item["id"] for item in response.data["results"]]
        self.assertIn(seen_notification.id, seen_ids)
        self.assertNotIn(unseen_notification.id, seen_ids)

        response = self.client.get(self.list_url, {"seen": "false"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        unseen_ids = [item["id"] for item in response.data["results"]]
        self.assertIn(unseen_notification.id, unseen_ids)
        self.assertNotIn(seen_notification.id, unseen_ids)

    def test_filter_by_user(self):
        NotificationUserFactory(user=self.user, notification=self.notification)
        NotificationUserFactory(
            user=self.other_user, notification=self.other_notification
        )

        response = self.client.get(self.list_url, {"user": self.user.id})

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        user_ids = [item["user"] for item in response.data["results"]]
        self.assertIn(self.user.id, user_ids)
        self.assertNotIn(self.other_user.id, user_ids)

    def test_filter_by_notification(self):
        NotificationUserFactory(user=self.user, notification=self.notification)
        NotificationUserFactory(
            user=self.user, notification=self.other_notification
        )

        response = self.client.get(
            self.list_url, {"notification": self.notification.id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        notification_ids = [
            item["notification"] for item in response.data["results"]
        ]
        self.assertIn(self.notification.id, notification_ids)
        self.assertNotIn(self.other_notification.id, notification_ids)

    def test_filter_by_notification_kind(self):
        success_notification = NotificationFactory(
            kind=NotificationKindEnum.SUCCESS
        )
        warning_notification = NotificationFactory(
            kind=NotificationKindEnum.WARNING
        )

        success_nu = NotificationUserFactory(
            user=self.user, notification=success_notification
        )
        warning_nu = NotificationUserFactory(
            user=self.user, notification=warning_notification
        )

        response = self.client.get(
            self.list_url,
            {"notification_kind": NotificationKindEnum.SUCCESS.value},
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        result_ids = [item["id"] for item in response.data["results"]]
        self.assertIn(success_nu.id, result_ids)
        self.assertNotIn(warning_nu.id, result_ids)

    def test_filter_by_date_range(self):
        old_date = timezone.now() - timedelta(days=10)
        recent_date = timezone.now() - timedelta(days=1)

        old_notification = NotificationUserFactory(
            user=self.user, notification=self.notification
        )
        old_notification.created_at = old_date
        old_notification.save()

        recent_notification = NotificationUserFactory(
            user=self.user, notification=self.other_notification
        )
        recent_notification.created_at = recent_date
        recent_notification.save()

        filter_date = timezone.now() - timedelta(days=5)
        response = self.client.get(
            self.list_url, {"created_after": filter_date.isoformat()}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        result_ids = [item["id"] for item in response.data["results"]]
        self.assertIn(recent_notification.id, result_ids)
        self.assertNotIn(old_notification.id, result_ids)

    def test_filter_has_seen_at(self):
        seen_notification = NotificationUserFactory(
            user=self.user,
            notification=self.notification,
            seen=True,
            seen_at=timezone.now(),
        )
        unseen_notification = NotificationUserFactory(
            user=self.user,
            notification=self.other_notification,
            seen=False,
            seen_at=None,
        )

        response = self.client.get(self.list_url, {"has_seen_at": "true"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        result_ids = [item["id"] for item in response.data["results"]]
        self.assertIn(seen_notification.id, result_ids)
        self.assertNotIn(unseen_notification.id, result_ids)

        response = self.client.get(self.list_url, {"has_seen_at": "false"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        result_ids = [item["id"] for item in response.data["results"]]
        self.assertIn(unseen_notification.id, result_ids)
        self.assertNotIn(seen_notification.id, result_ids)

    def test_ordering_by_created_at_desc(self):
        NotificationUserFactory(user=self.user, notification=self.notification)
        NotificationUserFactory(
            user=self.user, notification=self.other_notification
        )

        response = self.client.get(self.list_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        if len(response.data["results"]) >= 2:
            first_item = response.data["results"][0]
            second_item = response.data["results"][1]
            self.assertGreaterEqual(
                first_item["created_at"], second_item["created_at"]
            )

    def test_ordering_by_seen(self):
        NotificationUserFactory(
            user=self.user, notification=self.notification, seen=True
        )
        NotificationUserFactory(
            user=self.user, notification=self.other_notification, seen=False
        )

        response = self.client.get(self.list_url, {"ordering": "seen"})

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertGreaterEqual(len(response.data["results"]), 2)

    def test_search_by_user_name(self):
        user_with_unique_name = UserAccountFactory(first_name="UniqueFirstName")
        NotificationUserFactory(
            user=user_with_unique_name, notification=self.notification
        )

        response = self.client.get(self.list_url, {"search": "UniqueFirstName"})

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_search_by_notification_title(self):
        unique_notification = NotificationFactory()
        unique_notification.title = "UniqueNotificationTitle"
        unique_notification.save()

        NotificationUserFactory(
            user=self.user, notification=unique_notification
        )

        response = self.client.get(
            self.list_url, {"search": "UniqueNotificationTitle"}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_unseen_count_with_unseen_notifications(self):
        self.client.force_authenticate(user=self.user)

        NotificationUserFactory(
            user=self.user, notification=self.notification, seen=False
        )
        NotificationUserFactory(
            user=self.user, notification=self.other_notification, seen=True
        )

        url = reverse("notification-user-unseen-count")
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 1)

    def test_unseen_count_no_unseen_notifications(self):
        self.client.force_authenticate(user=self.user)

        NotificationUserFactory(
            user=self.user, notification=self.notification, seen=True
        )

        url = reverse("notification-user-unseen-count")
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("count", response.data)

    def test_unseen_count_unauthenticated(self):
        url = reverse("notification-user-unseen-count")
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_mark_all_as_seen(self):
        self.client.force_authenticate(user=self.user)

        NotificationUserFactory(
            user=self.user, notification=self.notification, seen=False
        )
        NotificationUserFactory(
            user=self.user, notification=self.other_notification, seen=False
        )

        url = reverse("notification-user-mark-all-as-seen")
        response = self.client.post(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["success"], True)

        self.assertEqual(
            NotificationUser.objects.filter(user=self.user, seen=False).count(),
            0,
        )

    def test_mark_all_as_unseen(self):
        self.client.force_authenticate(user=self.user)

        NotificationUserFactory(
            user=self.user, notification=self.notification, seen=True
        )
        NotificationUserFactory(
            user=self.user, notification=self.other_notification, seen=True
        )

        url = reverse("notification-user-mark-all-as-unseen")
        response = self.client.post(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["success"], True)

        self.assertEqual(
            NotificationUser.objects.filter(user=self.user, seen=True).count(),
            0,
        )

    def test_mark_as_seen_specific(self):
        self.client.force_authenticate(user=self.user)

        nu1 = NotificationUserFactory(
            user=self.user, notification=self.notification, seen=False
        )
        nu2 = NotificationUserFactory(
            user=self.user, notification=self.other_notification, seen=False
        )

        url = reverse("notification-user-mark-as-seen")
        data = {"notification_user_ids": [nu1.id]}
        response = self.client.post(url, data, format="json")
        print("======= response =========", response.data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["success"], True)

        nu1.refresh_from_db()
        nu2.refresh_from_db()
        self.assertTrue(nu1.seen)
        self.assertFalse(nu2.seen)

    def test_mark_as_unseen_specific(self):
        self.client.force_authenticate(user=self.user)

        nu1 = NotificationUserFactory(
            user=self.user, notification=self.notification, seen=True
        )
        nu2 = NotificationUserFactory(
            user=self.user, notification=self.other_notification, seen=True
        )

        url = reverse("notification-user-mark-as-unseen")
        data = {"notification_user_ids": [nu1.id]}
        response = self.client.post(url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["success"], True)

        nu1.refresh_from_db()
        nu2.refresh_from_db()
        self.assertFalse(nu1.seen)
        self.assertTrue(nu2.seen)

    def test_mark_as_seen_empty_list(self):
        self.client.force_authenticate(user=self.user)

        url = reverse("notification-user-mark-as-seen")
        data = {"notification_user_ids": []}
        response = self.client.post(url, data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_validation_errors(self):
        response = self.client.post(self.list_url, {})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        data = {
            "user": self.user.id,
            "notification": 999999,
            "seen": False,
        }
        response = self.client.post(self.list_url, data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)



import pytest
from unittest.mock import Mock, patch

from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.test import RequestFactory

from tag.admin import (
    TagAdmin,
    TaggedItemAdmin,
    TagStatusFilter,
    TagUsageFilter,
    ContentTypeFilter,
    TagInLine,
)
from tag.factories import TagFactory
from tag.factories.tagged_item import TaggedProductFactory
from tag.models import Tag, TaggedItem
from product.factories import ProductFactory

User = get_user_model()


@pytest.fixture
def admin_request():
    factory = RequestFactory()
    request = factory.get("/admin/tag/tag/")
    request.user = Mock()
    request.user.is_authenticated = True
    request.user.is_staff = True
    request.user.is_superuser = True
    return request


@pytest.fixture
def tag_admin():
    return TagAdmin(Tag, AdminSite())


@pytest.fixture
def tagged_item_admin():
    return TaggedItemAdmin(TaggedItem, AdminSite())


@pytest.fixture
def tag_inline():
    return TagInLine(Tag, AdminSite())


@pytest.mark.django_db
class TestTagStatusFilter:
    def test_filter_lookups(self, admin_request):
        filter_instance = TagStatusFilter(
            admin_request, {}, Tag, TagAdmin(Tag, AdminSite())
        )

        lookups = filter_instance.lookups(
            admin_request, TagAdmin(Tag, AdminSite())
        )

        expected_lookups = [
            "active",
            "inactive",
            "used",
            "unused",
            "popular",
            "recent",
        ]
        lookup_values = [lookup[0] for lookup in lookups]

        for expected in expected_lookups:
            assert expected in lookup_values

    def test_filter_queryset_no_value(self, admin_request):
        TagFactory(active=True)
        TagFactory(active=False)

        filter_instance = TagStatusFilter(
            admin_request, {}, Tag, TagAdmin(Tag, AdminSite())
        )

        queryset = Tag.objects.all()
        filtered_queryset = filter_instance.queryset(admin_request, queryset)

        assert filtered_queryset.count() == 2


@pytest.mark.django_db
class TestTagUsageFilter:
    def test_filter_lookups(self, admin_request):
        filter_instance = TagUsageFilter(
            admin_request, {}, Tag, TagAdmin(Tag, AdminSite())
        )

        lookups = filter_instance.lookups(
            admin_request, TagAdmin(Tag, AdminSite())
        )

        expected_lookups = ["single", "low", "medium", "high", "very_high"]
        lookup_values = [lookup[0] for lookup in lookups]

        for expected in expected_lookups:
            assert expected in lookup_values

    def test_filter_queryset_no_value(self, admin_request):
        TagFactory()
        TagFactory()

        filter_instance = TagUsageFilter(
            admin_request, {}, Tag, TagAdmin(Tag, AdminSite())
        )

        queryset = Tag.objects.all()
        filtered_queryset = filter_instance.queryset(admin_request, queryset)

        assert filtered_queryset.count() == 2


@pytest.mark.django_db
class TestContentTypeFilter:
    def test_content_type_filter_lookups(self, admin_request):
        tag = TagFactory()
        product = ProductFactory()
        TaggedProductFactory(tag=tag, content_object=product)

        filter_instance = ContentTypeFilter(
            admin_request, {}, Tag, TagAdmin(Tag, AdminSite())
        )

        lookups = filter_instance.lookups(
            admin_request, TagAdmin(Tag, AdminSite())
        )

        assert len(lookups) > 0
        content_type_ids = [lookup[0] for lookup in lookups]
        product_ct = ContentType.objects.get_for_model(
            ProductFactory._meta.model
        )
        assert product_ct.id in content_type_ids

    def test_filter_queryset_no_value(self, admin_request):
        TagFactory()
        TagFactory()

        filter_instance = ContentTypeFilter(
            admin_request, {}, Tag, TagAdmin(Tag, AdminSite())
        )

        queryset = Tag.objects.all()
        filtered_queryset = filter_instance.queryset(admin_request, queryset)

        assert filtered_queryset.count() == 2


@pytest.mark.django_db
class TestTagAdmin:
    def test_tag_info_display(self, tag_admin):
        tag = TagFactory()

        result = tag_admin.tag_info(tag)

        assert str(tag.id) in result
        assert tag.label in result
        assert "font-medium" in result

    def test_status_badge_active(self, tag_admin):
        tag = TagFactory(active=True)

        result = tag_admin.status_badge(tag)

        assert "Active" in result
        assert "bg-green-50" in result

    def test_status_badge_inactive(self, tag_admin):
        tag = TagFactory(active=False)

        result = tag_admin.status_badge(tag)

        assert "Inactive" in result
        assert "bg-red-50" in result

    def test_usage_stats(self, tag_admin):
        tag = TagFactory()

        for _ in range(5):
            product = ProductFactory()
            TaggedProductFactory(tag=tag, content_object=product)

        result = tag_admin.usage_stats(tag)

        assert "5" in result
        assert "uses" in result

    def test_content_distribution(self, tag_admin):
        tag = TagFactory()

        for _ in range(3):
            product = ProductFactory()
            TaggedProductFactory(tag=tag, content_object=product)

        result = tag_admin.content_distribution(tag)

        assert "types" in result
        assert "Product" in result

    def test_sort_display(self, tag_admin):
        tag = TagFactory()
        Tag.objects.filter(id=tag.id).update(sort_order=100)
        tag.refresh_from_db()

        result = tag_admin.sort_display(tag)

        assert "100" in result
        assert "#100" in result

    def test_created_display(self, tag_admin):
        tag = TagFactory()

        result = tag_admin.created_display(tag)

        assert len(result) > 10
        assert "2025" in result or "2024" in result

    def test_tag_analytics(self, tag_admin):
        tag = TagFactory()

        result = tag_admin.tag_analytics(tag)

        assert "Label Length" in result
        assert "Word Count" in result
        assert "Active Status" in result

    def test_usage_analytics(self, tag_admin):
        tag = TagFactory()

        result = tag_admin.usage_analytics(tag)

        assert "Total Usage" in result
        assert "Recent Usage" in result

    def test_content_analytics(self, tag_admin):
        tag = TagFactory()

        result = tag_admin.content_analytics(tag)

        assert "Diversity Score" in result
        assert "Most Used Type" in result

    @patch.object(TagAdmin, "message_user")
    def test_activate_tags_action(
        self, mock_message_user, tag_admin, admin_request
    ):
        tag1 = TagFactory(active=False)
        tag2 = TagFactory(active=False)
        queryset = Tag.objects.filter(id__in=[tag1.id, tag2.id])

        tag_admin.activate_tags(admin_request, queryset)

        tag1.refresh_from_db()
        tag2.refresh_from_db()
        assert tag1.active is True
        assert tag2.active is True
        mock_message_user.assert_called_once()

    @patch.object(TagAdmin, "message_user")
    def test_deactivate_tags_action(
        self, mock_message_user, tag_admin, admin_request
    ):
        tag1 = TagFactory(active=True)
        tag2 = TagFactory(active=True)
        queryset = Tag.objects.filter(id__in=[tag1.id, tag2.id])

        tag_admin.deactivate_tags(admin_request, queryset)

        tag1.refresh_from_db()
        tag2.refresh_from_db()
        assert tag1.active is False
        assert tag2.active is False
        mock_message_user.assert_called_once()

    @patch.object(TagAdmin, "message_user")
    def test_update_sort_order_action(
        self, mock_message_user, tag_admin, admin_request
    ):
        tag1 = TagFactory()
        tag2 = TagFactory()
        queryset = Tag.objects.filter(id__in=[tag1.id, tag2.id])

        tag_admin.update_sort_order(admin_request, queryset)

        mock_message_user.assert_called_once()

    @patch.object(TagAdmin, "message_user")
    def test_analyze_usage_action(
        self, mock_message_user, tag_admin, admin_request
    ):
        tag1 = TagFactory()
        tag2 = TagFactory()
        queryset = Tag.objects.filter(id__in=[tag1.id, tag2.id])

        tag_admin.analyze_usage(admin_request, queryset)

        mock_message_user.assert_called_once()


@pytest.mark.django_db
class TestTaggedItemAdmin:
    def test_tagged_item_info_display(self, tagged_item_admin):
        tag = TagFactory()
        product = ProductFactory()
        tagged_item = TaggedProductFactory(tag=tag, content_object=product)

        result = tagged_item_admin.tagged_item_info(tagged_item)

        assert str(tagged_item.id) in result
        assert "font-medium" in result

    def test_tag_display(self, tagged_item_admin):
        tag = TagFactory(active=True)
        product = ProductFactory()
        tagged_item = TaggedProductFactory(tag=tag, content_object=product)

        result = tagged_item_admin.tag_display(tagged_item)

        assert tag.label in result
        assert any(
            color in result
            for color in [
                "bg-blue-50",
                "bg-green-50",
                "text-blue-600",
                "text-green-600",
            ]
        )

    def test_content_object_display(self, tagged_item_admin):
        tag = TagFactory()
        product = ProductFactory()
        tagged_item = TaggedProductFactory(tag=tag, content_object=product)

        result = tagged_item_admin.content_object_display(tagged_item)

        assert str(product) in result
        assert "ID:" in result

    def test_content_type_badge(self, tagged_item_admin):
        tag = TagFactory()
        product = ProductFactory()
        tagged_item = TaggedProductFactory(tag=tag, content_object=product)

        result = tagged_item_admin.content_type_badge(tagged_item)

        assert "Product" in result
        assert "bg-green-50" in result

    def test_created_display(self, tagged_item_admin):
        tag = TagFactory()
        product = ProductFactory()
        tagged_item = TaggedProductFactory(tag=tag, content_object=product)

        result = tagged_item_admin.created_display(tagged_item)

        assert len(result) > 10
        assert "2025" in result or "2024" in result


@pytest.mark.django_db
class TestTagInLine:
    def test_tag_inline_fields(self, tag_inline):
        assert tag_inline.fields == ("tag",)
        assert tag_inline.extra == 0
        assert tag_inline.verbose_name == "Tag"
        assert tag_inline.verbose_name_plural == "Tags"

    def test_tag_inline_model(self, tag_inline):
        assert tag_inline.model == TaggedItem


@pytest.mark.django_db
class TestTagAdminIntegration:
    def test_get_queryset_annotations(self, tag_admin, admin_request):
        tag = TagFactory()
        product = ProductFactory()
        TaggedProductFactory(tag=tag, content_object=product)

        queryset = tag_admin.get_queryset(admin_request)
        tag_obj = queryset.get(id=tag.id)

        assert hasattr(tag_obj, "usage_count")
        assert tag_obj.usage_count == 1

        assert hasattr(tag_obj, "content_types_count")
        assert tag_obj.content_types_count == 1

    def test_admin_configuration(self, tag_admin):
        assert tag_admin.list_per_page == 50
        assert "tag_info" in tag_admin.list_display
        assert "status_badge" in tag_admin.list_display
        assert "usage_stats" in tag_admin.list_display

        action_names = []
        for action in tag_admin.actions:
            if hasattr(action, "__name__"):
                action_names.append(action.__name__)
            else:
                action_names.append(str(action))

        assert "activate_tags" in action_names
        assert "deactivate_tags" in action_names
